//
//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// This source code is an unpublished proprietary document, and is available only
// under a written license.
// 

#include <sys/types.h>

#if !defined(__AGCFILTERS__H__)
#define __AGCFILTERS__H__ 1
//#include <math.h>

#define FIRARRAYTYPE static const

#include "itqA.h"
#include "varray.h"
#include "vecmath.h"
#include "accuminfo.h"

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wconversion"

#include "coeff-96000-255fr20300lp.h"

#pragma clang diagnostic pop

#if !defined(__clang__)
#define AGCFNOINLINE //__attribute__((noinline))
#else
#define AGCFNOINLINE //__attribute__((noinline))
#endif

template <typename VDEF>
inline varrayT<VDEF>
sratetmod(varrayT<VDEF> tval)
{
	return Vselect((tval == varrayT<VDEF>(1.0)), varrayT<VDEF>(1.0), tval);
}

template <typename T>
inline T
sratetmod(T tval)
{
	return (tval == T(1.0)) ? T(1.0) : (tval * T(aspec.SRATETMULT));
}

#define THQALIGNED

#include "THQbase.h"

struct audiopayload
{
	int samplecount;
	Asample samples[AUDIOCHUNKSIZE];
	audiopayload(void) : samplecount(), samples() {}
};

typedef THQpayload <audiopayload> audiochunk;
typedef audiochunk *paudiochunk;

template <typename T, typename CT = double>
class hpfT
{
	T coeff;
	T zval;

public:
	hpfT(void) : coeff(), zval() {}

	hpfT(const hpfT<T, CT>& init) : coeff(init.coeff), zval(init.zval) {}

	hpfT(T c, T z = T{}): coeff(c), zval(z) {}
	
	hpfT(CT c, CT z = CT(0)) : hpfT(T(c), T(z)) {}

	hpfT<T, CT>&
	operator=(hpfT<T, CT> src)
	{
		coeff = src.coeff;
		zval = src.zval;
		return *this;
	}

	T
	next(T in)
	{
		T out = in - zval;

		zval = zval + out * coeff;

		return out;
	}

	T
	operator()(T in)
	{
		return next(in);
	}

	void
	setcoeff(T newcoeff)
	{
		coeff = newcoeff;
	}

	void
	setcoeff(CT newcoeff)
	{
		coeff = T(newcoeff);
	}
};

class hpfAs
{
	float coeff;
	Asample zval;

public:
	hpfAs(float c) :
		coeff(c), zval(0.0f, 0.0f)
	{}

	Asample
	next(Asample in)
	{
		Asample out = in - zval;

		zval = zval + out * coeff;

		return out;
	}

	Asample
	operator()(const Asample& in)
	{
		return next(in);
	}
};

class hpf4As
{
	float coeff[4];
	Asample zval[4];

public:

	hpf4As(float c0, float c1, float c2, float c3) :
		coeff(), zval()
	{
		for (int i = 0; i < 4; i++)
		{
			zval[i] = Asample{};
		}
		coeff[0] = c0;
		coeff[1] = c1;
		coeff[2] = c2;
		coeff[3] = c3;
	}

	hpf4As(float c0) :
		hpf4As(c0, c0, c0, c0) {}

	void
	setcoeffs(float c0, float c1, float c2, float c3)
	{
		coeff[0] = c0;
		coeff[1] = c1;
		coeff[2] = c2;
		coeff[3] = c3;
	}
		
	void
	next(const Asample* in, Asample* out)
	{
		for (int i = 0; i < 4; i++)
		{
			out[i] = in[i] - zval[i];
			zval[i] = zval[i] + out[i] * coeff[i];
		}
		return;
	}

	void
	operator()(const Asample* in, Asample* out)
	{
		next(in, out);
		return;
	}
};

template <int NFLT>
class sfdnfAs GCQ
{
	int inptr;
	float rawbuf0[NFLT*2];
	float rawbuf1[NFLT*2];
	const float * filter;

	void
	newdata(const Asample& in)
	{
		float tmp = in.item0();
		rawbuf0[inptr] = tmp;
		rawbuf0[inptr+NFLT] = tmp;

		tmp = in.item1();
		rawbuf1[inptr] = tmp;
		rawbuf1[inptr+NFLT] = tmp;
		inptr += 1;
		if (inptr >= NFLT)
			inptr = 0;
	}

	Asample
	basicfilt(void) const
	{
		
		float out0 = 0.0f, out1 = 0.0f;
		const float * rbptr0 = &rawbuf0[inptr];
		const float * rbptr1 = &rawbuf1[inptr];
		for (int i = 0; i < NFLT; i++)
		{
			out0 += rbptr0[i] * filter[i];
			out1 += rbptr1[i] * filter[i];
		}
		return Asample{out0, out1};
	}


public:
	sfdnfAs(const float *filtera) NOINLINE : inptr(0), rawbuf0(), rawbuf1(), filter(filtera) 
	{
		for (int i = 0; i < NFLT*2; i++)
		{
			rawbuf0[i] = 0.0f;
			rawbuf1[i] = 0.0f;
		}
	}

	void setfilt(const float *filtera)
	{
		filter = filtera;
	}

	Asample NOINLINE
	next(const Asample& input)
	{
		newdata(input);
		return basicfilt();
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

};

template <int ORDER>
class sfIIRAs GCQ
{
	int inptr;
	Asample Fbuf[(ORDER+1)*2];
	Asample Ibuf[(ORDER+1)*2];

	float Ffilter[ORDER+1];
	float Ifilter[ORDER];

	Asample __attribute__((always_inline))
	basicfilt(const Asample& in) 
	{
		Fbuf[inptr] = in;
		Fbuf[inptr+(ORDER + 1)] = in;
		
		Asample out{};

#if 0
		const Asample * Fbufptr = &Fbuf[inptr + (ORDER + 1)];
		for (int i = ORDER; i >= 0; --i)
		{
			out += *(Fbufptr - i) * Ffilter[(ORDER - i)];
		}

		const Asample * Ibufptr = &Ibuf[inptr + (ORDER + 1)];
		for (int i = ORDER; i > 0; --i)
		{
			out += *(Ibufptr - i) * Ifilter[(ORDER - i)];
		}
#else
		const Asample * Fbufptr = &Fbuf[inptr + (ORDER + 1)];
		for (int i = ORDER; i >= 0; --i)
		{
			float fc = Ffilter[(ORDER - i)];
			out.vec() += (Fbufptr - i)->vec() * v2sf{fc,fc};
		}

		const Asample * Ibufptr = &Ibuf[inptr + (ORDER + 1)];
		for (int i = ORDER; i > 0; --i)
		{
			float fc = Ifilter[(ORDER - i)];
			out.vec() += (Ibufptr - i)->vec() * v2sf{fc,fc};
		}

#endif

		Ibuf[inptr] = out;
		Ibuf[inptr+(ORDER + 1)] = out;

		inptr += 1;
		if (inptr >= (ORDER + 1))
			inptr = 0;
			
		return out;
	}


public:
	sfIIRAs(const float *Ff, const float *If) NOINLINE : inptr(0), Fbuf(), Ibuf(), Ffilter(), Ifilter()
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
		for (int i = 0; i < (ORDER + 1)*2; i++)
		{
			Fbuf[i] = Asample{};
			Ibuf[i] = Asample{};
		}
	}

	void
	setfilt(const float *Ff, const float *If)
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
	}

	Asample __attribute__((always_inline))
	next(const Asample& input)
	{
		return basicfilt(input);
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

};

template <int ORDER, typename DT, typename CT = double>
class sfIIRT GCQ
{
	int inptr;
	DT Fbuf[(ORDER+1)*2];
	DT Ibuf[(ORDER+1)*2];

	CT Ffilter[ORDER+1];
	CT Ifilter[ORDER];

	DT
	basicfilt(DT in) 
	{
		Fbuf[inptr] = in;
		Fbuf[inptr+(ORDER + 1)] = in;
		
		DT out{};

		const DT * Fbufptr = &Fbuf[inptr + (ORDER + 1)];
		for (int i = ORDER; i >= 0; --i)
		{
			out += *(Fbufptr - i) * Ffilter[(ORDER - i)];
		}

		const DT * Ibufptr = &Ibuf[inptr + (ORDER + 1)];
		for (int i = ORDER; i > 0; --i)
		{
			out += *(Ibufptr - i) * Ifilter[(ORDER - i)];
		}

		Ibuf[inptr] = out;
		Ibuf[inptr+(ORDER + 1)] = out;

		inptr += 1;
		if (inptr >= (ORDER + 1))
			inptr = 0;
			
		return out;
	}


public:
	sfIIRT(const CT *Ff, const CT *If) : inptr(0), Fbuf(), Ibuf(), Ffilter(), Ifilter()
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
		for (int i = 0; i < (ORDER + 1)*2; i++)
		{
			Fbuf[i] = DT{};
			Ibuf[i] = DT{};
		}
	}

	sfIIRT(const float *Ff, const float *If) : inptr(0), Fbuf(), Ibuf(), Ffilter(), Ifilter()
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
		for (int i = 0; i < (ORDER + 1)*2; i++)
		{
			Fbuf[i] = DT{};
			Ibuf[i] = DT{};
		}
	}

	void
	setfilt(const CT *Ff, const CT *If)
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
	}

	void
	setfilt(const float *Ff, const float *If)
	{
		for (int i = 0; i < (ORDER + 1); i++)
			Ffilter[i] = Ff[i];
		for (int i = 0; i < ORDER; i++)
			Ifilter[i] = -1.0f * If[i];
	}

	DT next(DT input)
	{
		return basicfilt(input);
	}

	DT operator()(DT input)
	{
		return next(input);
	}

};

template <int NDLY>
class sdlyAs GCQ
{
	int inptr;
	Asample rawbuf[NDLY];

	void
	newdata(const Asample& in)
	{
		rawbuf[inptr] = in;
		inptr += 1;
		if (inptr >= NDLY)
			inptr = 0;
	}

public:
	sdlyAs() : inptr(0), rawbuf()
	{
		for (int i = 0; i < NDLY; i++)
		{
			rawbuf[i] = Asample(0.0f,0.0f);
		}
	}
	const Asample // NOINLINE
	next(const Asample& in)
	{
		newdata(in);
		return rawbuf[inptr];
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

};

template <int MAXDLY>
class sVdlyAs GCQ
{
	int inptr;
	Asample rawbuf[2*MAXDLY];

	void
	newdata(const Asample& in)
	{
	//	::memcpy(&rawbuf[inptr], &in, sizeof(Asample));
	//	::memcpy(&rawbuf[inptr+MAXDLY], &in, sizeof(Asample));
		rawbuf[inptr] = in;
		rawbuf[inptr+MAXDLY] = in;
		inptr += 1;
		if (inptr >= MAXDLY)
			inptr = 0;
	}

public:
	sVdlyAs() : inptr(0), rawbuf()
	{
#pragma clang loop unroll(disable)
		for (int i = 0; i < 2*MAXDLY; i++)
		{
			rawbuf[i] = Asample(0.0f,0.0f);
		}
	}

	const Asample&
	next(const Asample& in)
	{
		newdata(in);
		return rawbuf[inptr];
	}

	const Asample& 
	next(const Asample& in, int vdly)
	{
		newdata(in);
		return rawbuf[inptr + (MAXDLY-1) - vdly];
	}

	const Asample&
	dly(int dlyamt) const
	{
		return rawbuf[inptr + (MAXDLY-1) - dlyamt];
	}

	const Asample& operator()(const Asample& input)
	{
		return next(input);
	}

	const Asample& operator()(const Asample& input, int delay)
	{
		return next(input, delay);
	}

	const Asample& operator()(int dlyamt) const
	{
		return dly(dlyamt);
	}

};

template <int MAXDLY>
class sVdly8As GCQ
{
	struct asa
	{
		Asample d[8];
		asa(void)
		{
			for (int i = 0; i < 8; i++)
				d[i] = Asample(0.0f,0.0f);
		}
	};

	asa rawbuf[2*MAXDLY];
	int inptr;

	void __attribute__((always_inline))
	newdata(const Asample* in)
	{
		Asample *asp0 = &rawbuf[inptr].d[0];
		Asample *asp1 = &rawbuf[inptr+MAXDLY].d[0];

		for (int i = 0; i < 8; i++)
		{
			asp0[i].vec() = in[i].vec();
			asp1[i].vec() = in[i].vec();
		}

		inptr += 1;
		if (inptr >= MAXDLY)
			inptr = 0;
	}

public:
	sVdly8As() : rawbuf(), inptr(0)
	{
	}

	const Asample* __attribute__((always_inline))
	next(const Asample* in)
	{
		newdata(in);
		return rawbuf[inptr].d;
	}

	const Asample*
	dly(int dlyamt) const
	{
		return rawbuf[inptr + ((MAXDLY-1) - dlyamt)].d;
	}

	const Asample* operator()(const Asample* input)
	{
		return next(input);
	}

	const Asample* operator()(int dlyamt) const
	{
		return dly(dlyamt);
	}

};


namespace dbyfilt
{
	static constexpr float fir80lp_192[3] = {1.710305e-6, 3.4206117e-6, 1.710305e-6};
	static constexpr float iir80lp_192[2] = {9.9630444e-1, -1.9962976};
	static constexpr float fir80lp_96[3] = {6.828594e-6, 1.3657189e-5, 6.828594e-6};
	static constexpr float iir80lp_96[2] = {0.9926225, -1.9925952};
	static constexpr float fir80lp_48[3] = {2.7213808e-5, 5.4427616e-5, 2.7213808e-5};
	static constexpr float iir80lp_48[2] = {0.9852995, -1.9851907};
	static constexpr float fir80lp_441[3] = {3.2218972e-5, 6.4437944e-5, 3.2218972e-5};
	static constexpr float iir80lp_441[2] = {0.984009917, -1.9838810};
	static constexpr float fir3000hp_192[3] = {0.93293216, -1.86586431, 0.93293216};
	static constexpr float iir3000hp_192[2] = {0.8703674775, -1.861361147};
	static constexpr float fir3000hp_96[3] = {0.8703308, -1.7406616, 0.8703308};
	static constexpr float iir3000hp_96[2] = {0.7575469, -1.7237762};
	static constexpr float fir3000hp_48[3] = {0.757076375, -1.51415275, 0.757076375};
	static constexpr float iir3000hp_48[2] = {0.5740619151, -1.454243586};
	static constexpr float fir3000hp_441[3] = {0.7385387094, -1.477077419, 0.7385387094};
	static constexpr float iir3000hp_441[2] = {0.5466494937, -1.407505344};
	static constexpr float fir9000hp_192[3] = {0.8118317459, -1.623663492, 0.8118317459};
	static constexpr float iir9000hp_192[2] = {0.6493898773, -1.587937106};
	static constexpr float fir9000hp_96[3] = {0.6574552, -1.3149104, 0.6574552};
	static constexpr float iir9000hp_96[2] = {0.4359074, -1.19391336};
	static constexpr float fir9000hp_48[3] = {0.418163346, -0.8363266915, 0.418163346};
	static constexpr float iir9000hp_48[2] = {0.2097153578, -0.4629380253};
	static constexpr float fir9000hp_441[3] = {0.3827827894, -0.765565579, 0.3827827894};
	static constexpr float iir9000hp_441[2] = {0.1919799727, -0.3391511849};

//
// output 20500Hz, 96k lpf, 4th order
//
//	static constexpr float Fiir20500lp4_96[5] = {0.1384310, 0.2939733, 0.3958198, 0.2939733, 0.138431};
//	static constexpr float Iiir20500lp4_96[4] = {0.1128558, -0.2656563, 0.9360753, -0.5099126};
//
// output 20500Hz, 96k lpf, 4th order
//
	//
	// elliptical -- 20500 cutoff, [0.005, 0.01]
	//
	static constexpr float Fiir20500lp4_192[5] = {0.0310378287, 0.0190729332, 0.04662629242, 0.0190729332, 0.03103782871};
	static constexpr float Iiir20500lp4_192[4] = {0.2481306858, -1.186241625, 2.353466675, -2.26776999};

	//
	// elliptical -- 20500 cutoff, [0.005, 0.01]
	//
	static constexpr float Fiir20500lp4_96[5] = {0.1157991894, 0.3285279539, 0.4493870996, 0.3285279539, 0.1157991894};
	static constexpr float Iiir20500lp4_96[4] = {0.07850221322, -0.1854407589, 0.7903590362, -0.338655278};

	//
	// elliptical -- 20500 cutoff, [0.005, 0.01]
	//
	static constexpr float Fiir20500lp4_48[5] = {0.6255939943, 2.471228746, 3.691476826, 2.471228746, 0.6255939943};
	static constexpr float Iiir20500lp4_48[4] = {0.3990939381, 1.924339230, 3.578670373, 3.032692748};

	//
	// elliptical -- 20500 cutoff, [0.005, 0.01]
	//
	static constexpr float Fiir20500lp4_441[5] = {0.7973545529, 3.180378648, 4.766061830, 3.180378648, 0.7973545529};
	static constexpr float Iiir20500lp4_441[4] = {0.643351724, 2.845839863, 4.750057959, 3.546205935};
//
// output 20500Hz, 96k lpf, 3rh order
//
// output 20500Hz, 96k lpf, 3rh order
//
	static constexpr float Fiir20500lp3_96[4] = {0.2771420, 0.820992, 0.820992, 0.2771420};
	static constexpr float Iiir20500lp3_96[3] = {0.0559860, 0.4901972, 0.6501048};
};

class lpf20500
{
	using FTYPE=sfIIRAs<4>;
	FTYPE lpf;
public:
	lpf20500() : lpf(dbyfilt::Fiir20500lp4_96, dbyfilt::Iiir20500lp4_96)
	{
		switch(aspec.inSRATE)
		{
	case 44100:
		lpf.setfilt(dbyfilt::Fiir20500lp4_441, dbyfilt::Iiir20500lp4_441);
		break;
	case 48000:
		lpf.setfilt(dbyfilt::Fiir20500lp4_48, dbyfilt::Iiir20500lp4_48);
		break;
	case 96000:
		lpf.setfilt(dbyfilt::Fiir20500lp4_96, dbyfilt::Iiir20500lp4_96);
		break;
	case 192000:
		lpf.setfilt(dbyfilt::Fiir20500lp4_192, dbyfilt::Iiir20500lp4_192);
		break;
		}
	}

	Asample __attribute__((always_inline)) next(const Asample& in)
	{
		return lpf.next(in);
	}

	Asample __attribute__((always_inline)) operator() (const Asample& in)
	{
		return next(in);
	}
};

class dATsingle
{
	using FTYPE=sfIIRAs<2>;

	FTYPE fr80lp;
	FTYPE fr3000hp;
	FTYPE fr9000hp;
	lpf20500 aaf96;

public:

	dATsingle(void) NOINLINE :
		fr80lp(dbyfilt::fir80lp_96, dbyfilt::iir80lp_96),
		fr3000hp(dbyfilt::fir3000hp_96, dbyfilt::iir3000hp_96),
		fr9000hp(dbyfilt::fir9000hp_96, dbyfilt::iir9000hp_96),
		aaf96()
	{
		switch(aspec.inSRATE)
		{
	case 44100:
		fr80lp.setfilt(dbyfilt::fir80lp_441, dbyfilt::iir80lp_441);
		fr3000hp.setfilt(dbyfilt::fir3000hp_441, dbyfilt::iir3000hp_441);
		fr9000hp.setfilt(dbyfilt::fir9000hp_441, dbyfilt::iir9000hp_441);
		break;
	case 48000:
		fr80lp.setfilt(dbyfilt::fir80lp_48, dbyfilt::iir80lp_48);
		fr3000hp.setfilt(dbyfilt::fir3000hp_48, dbyfilt::iir3000hp_48);
		fr9000hp.setfilt(dbyfilt::fir9000hp_48, dbyfilt::iir9000hp_48);
		break;
	case 96000:
		fr80lp.setfilt(dbyfilt::fir80lp_96, dbyfilt::iir80lp_96);
		fr3000hp.setfilt(dbyfilt::fir3000hp_96, dbyfilt::iir3000hp_96);
		fr9000hp.setfilt(dbyfilt::fir9000hp_96, dbyfilt::iir9000hp_96);
		break;
	case 192000:
		fr80lp.setfilt(dbyfilt::fir80lp_192, dbyfilt::iir80lp_192);
		fr3000hp.setfilt(dbyfilt::fir3000hp_192, dbyfilt::iir3000hp_192);
		fr9000hp.setfilt(dbyfilt::fir9000hp_192, dbyfilt::iir9000hp_192);
		break;
		}
	}

	void NOINLINE
	next( const Asample& in, Asample* out)
	{
		Asample fin = aaf96(in);
		Asample tmplp80 = fr80lp.next(fin);
		Asample tmphp3000 = fr3000hp.next(fin);
		Asample tmphp9000 = fr9000hp.next(fin);

		out[0] = tmplp80;
		out[1] = fin - (tmphp3000 + tmplp80);
		out[2] = tmphp3000;
		out[3] = tmphp9000;
	}

	void
	operator()(const Asample& in, Asample* out)
	{
		return next(in, out);
	}
};

#endif
//
// vim: set ts=4:
//

