//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// You can use this code under one of two licenses:
//
// This source code is an unpublished proprietary document, and is available 
// for commercial (money making) purposes only under a written license.  If you
// happen to get a copy of this document without agreement, then all you have
// to do is to adhere to the GPL below.
//
// Or,
//
// You can adhere to the GPL, and that will also be okay (in this case,
// no written license is otherwise needed.)  I am not in love with the
// GPL -- and don't mind if the ideas are reused, but prefer that
// my code not be directly used commercially without my knowedge.
// I want to know what is going on with the code.
//
// Code modifications are NOT accepted (unless there is a mutually signed
// written agreement.)  Otherwise, reports, and requests to fix bugs
// are happily received.  (this is for licensing reasons, not for ego reasons.)
//
// A written agreement can modify any part of these copyright/licensing terms
// above.  Such an agreement must be signed/counter-signed by both parties.
// 

//
// This file contains a suprisingly effective and simple audio expansion processor
//
//
// SOME OF THE CAPABILITY OF THIS CODE IS REDUNDANT -- DUE TO AN EARLIER BUG, I HAD THOUGHT
// THAT IT WAS NECESSARY TO RUN MULTI-AXIS decoding, which might still be helpful for audible
// quality -- but not necessary to decode the targeted compression.  Doing additional
// expansion would require optimized algorithms for that purpose, and such code does exist
// in my finalizing processor program (this code was partially derived from the finalizer.)
//
// THERE ARE LOTS OF DIFFERENCES BETWEEN THIS AN MY FINALIZING PROCESSOR.  The finalizer uses
// 8 bands of audio, and not just 4 like this expander.  It also uses the very advanced
// attack/decay code, and has multiple compression/expansion passes tuned to various purposes,
// with a distortion free hard limiter scheme.  Up to 5-6 of various audio processing passes,
// with no gain control intermod (or at least, very-very little!!!)  Some of the attack/release
// sections can be enabled to have FULL waveform following -- which does not sound ideal, but
// with a little additional modification of the gain control -- works IDEALLY.
//

//
// There is excess precision in the calculations here: vectors of double precision
// instead of using floats...  However, the compiler that I use sometimes chokes on 
// non-full-native vectors -- it sometimes produces worse code with vectors of 4 floats instead
// of vectors of 4 doubles.   Since I am doing very few divisions, the cost of doing
// doubles instead of floats on an Intel architecture isn't very significant, but the cost
// of garbled (but correct) code due to using short vectors of floats can be significant.
// None of the large storage buffers use double precision, so memory bandwidth is also
// not excessively impacted (maybe register spills, short term storage and minimal argument
// passing.)
//
// The finalizing processor just uses floats -- and I am considering a reconfiguration of the
// existing 8 bands -- possibly up to 16bands.  Rather than being approx octave width, they would
// be approx 1/2 octave -- that opens up some other interesting possibilities, but that
// is a rather aggressive move -- and might be more trouble than it is worth.  It is already
// tricky to properly handle the 8 bands without overly 'flattening' the spectrum.  I am
// using some good methods that allow the processor to maintain the same spectrum shape without
// creating 'policies' which might change the sound.
//
// HINTS ON READING THE VECTOR CODE:
// * the vector type vtype4 is a vector of 4 double precision floats
//
// * when doing a constructor or type cast that looks like 'val = vtype4(-1.0)', it
//       fills all vector members with the specified value.
//
// * when doing a constructor or type cast that looks like 'val = vtype4({a,b,c,d})',
//       it fills in each element as val[0] = a, val[1] = b, val[2] = c,
//       val[3] = d.
//
// * very important: the vector type descriptor is a type in its own right.  That
//       class type describes all aspects of the integer or float vector type
//       of the same shape.  For example, an vector of 4 32bit floating point
//       and a vector of 4 32bit integers use the same descriptor type.  The
//       descriptor describes the type sizes, the underlying types, etc.
//       The actual vector type is usually either varrayT<VDESC> or IvarrayT<VDESC>.
//       There is also a raw vector type like UvarrayT<VDESC> that remembers the
//       shape, but data that has that type can only do conversions or raw assignments.
//       "vtype4" is a typedef of varrayT<TYPEvd4x1>, which means 'a vector of
//       four doubles with a count of one such HW vector.
//
// * math between vectors of the same type are supported.  All of +,-,*,/,
//       and, or, xor, Vmax, Vmin, Vgt, Vlt, Vge, Vle, (Vselect -- does
//			an ?: type operation on each element of the vector),
//       pow(vec,scalar), pow(vec,vec), log, log10, exp, exp10, dB2lin, lin2dB,
//       ([] -- array access of each element), var.fprint(FILEP, fmt),
//       and probably other things.
//
// * some math between vectors and scalars is very partially supported -- I haven't
//      filled in the combinations that support all of the vector/scalar combos.
//      Multiplication that does 'vector *= scalar' does work, but not much more.
//
// * some math between vector types (e.g. vector of 4 doubles vs. vector of 4 floats) is
//      partially supported.  Assignments ARE supported between types, but are meant to be
//      warned/errored if misused -- so mistakes are attempted to be caught.
//
// * some math between vector types that make little sense almost works (e.g. addition
//      between a vector size of 4 and a vector size of 8.) Expect 'interesting'
//      results until this is debugged.
//
// * no attempt is made to support matrices.
//
// * there are integer analogs of all floating point vector types.  Math between such types
//      is limited due to lack of good semantic definition.
//
// * assignment between integer and floating point vector types does the full type conversion,
//      but no special cases are handled, and there is no unsigned integer type yet.
//
// * there is an idiom for RAW copying between integer and floating point types "intvec = intvectype(fpvec.raw())"
//      which devolves the fp vec type to a raw type (with only conversion operations), and then raw copies
//      to the integer.  There are limited purposes for this, but some of the HW vector
//      opcodes (vblendvps/vblendvpd) encourage a certain mix between integer and FP because of the
//      CPU/HW use of an fp value as a logical bitmask.
//
// * the vector code is designed to support multiple HW vectors as one logical one -- but
//      the multiple vectors aren't fully supported yet.  All of the structure does exist.
//


//
// All timings are for 44.1k -- there are some underlying support mechanisms that handle the conversions
// to other sample rates.
//
// For REAL fun, disassemble this turkey -- the vector code can make it almost impossible to trace without
// experience in working with the vector CPU operations. :-).
//

#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>

#include <ctype.h>

#include <malloc.h>
#include <signal.h>

//
// itqA.h contains manifest constants
//
#include "itqA.h"

//
// general typedefs
//
#include "itq-typedefs.h"

//
// inter-thread communicatsion queues/queue management (thread stuff is pretty much dead for
// the expansion processor -- it is too simple to require threading.)
//
//#include "itq-qdefs.h"

//
// miscellaneous filters (IIR/FIR filter code, filter coefficients/etc.)
//
#include "agcfilters.h"
#include "math.h"

//
// Stupid/poorly written low level audio I/O routines
//
#include "audioio.h"

//
// Data recording/gain logging/etc
//
#include "accuminfo.h"

//
// This is a filter class from the very powerful 'smfilt.h' file which contains all kinds of attack/release schemes.
// I have included it here so that the dolbyA code can be kept seperate so that it can be distributed without giving
// away my cool secrets. :-).
//
// So, this (VtavfiltDYN) is a simple nonlinear attack/fixed release time scheme.
//
#if 0
template <typename VDEF, int NEG = false>
class VtavfiltDYN GCQ
{
	typedef varrayT<VDEF> VT;
	typedef IvarrayT<VDEF> IVT;
	typedef typename VDEF::FT FT;
	typedef typename VDEF::IT IT;

	using VAFT=VT;
	using VAIT=IVT;

	VAFT curval;
	VAFT curtarget;
	VAFT filt;
	VAFT avfilt;
	VAFT av1filt;
	VAFT av2filt;
	VAFT av3filt;
	VAFT av4filt;

	typedef VtavfiltDYN<VDEF,NEG> me;

public:

	using bf=VAFT;
	using bi=VAIT;
	static constexpr int VS = VDEF::VS;

	VtavfiltDYN() : curval(), curtarget(), filt(), avfilt() {}

	VtavfiltDYN(bf fval, bf avfval, bf av1fval, bf av2fval, bf av3fval, bf av4fval, bf initval) :
		curval(initval), curtarget(initval), filt(fval),
			avfilt(avfval), av1filt(av1fval), av2filt(av2fval), av3filt(av3fval), av4filt(av4fval)
		{
		}

	VtavfiltDYN(bf fval, bf avfval, bf av1fval, bf av2fval, bf av3fval, bf initval) :
		curval(initval), curtarget(initval), filt(fval),
			avfilt(avfval), av1filt(av1fval), av2filt(av2fval), av3filt(av3fval), av4filt(bf(FT(0)))
		{
		}

	VtavfiltDYN(bf fval, bf avfval, bf av1fval, bf av2fval, bf initval) :
		curval(initval), curtarget(initval), filt(fval),
			avfilt(avfval), av1filt(av1fval), av2filt(av2fval), av3filt(bf(FT(0))), av4filt(bf(FT(0)))
		{
		}

	VtavfiltDYN(bf fval, bf avfval, bf av1fval, bf initval) :
		curval(initval), curtarget(initval), filt(fval),
			avfilt(avfval), av1filt(av1fval), av2filt(bf(FT(0))), av3filt(bf(FT(0))), av4filt(bf(FT(0)))
		{
		}

	VtavfiltDYN(bf fval, bf avfval, bf initval) :
		curval(initval), curtarget(initval), filt(fval),
			avfilt(avfval), av1filt(bf(FT(0))), av2filt(bf(FT(0))), av3filt(bf(FT(0))), av4filt(bf(FT(0)))
		{
		}

	VtavfiltDYN(FT fval, FT avfval, FT av1fval, FT av2fval, FT av3fval, FT av4fval,  FT initval) :
		VtavfiltDYN(VT(fval), VT(avfval), VT(av1fval), VT(av2fval), VT(av3fval), VT(av4fval), VT(initval))
		{
		}

	VtavfiltDYN(FT fval, FT avfval, FT av1fval, FT av2fval, FT av3fval, FT initval) :
		VtavfiltDYN(VT(fval), VT(avfval), VT(av1fval), VT(av2fval), VT(av3fval), VT(initval))
		{
		}

	VtavfiltDYN(FT fval, FT avfval, FT av1fval, FT av2fval, FT initval) :
		VtavfiltDYN(VT(fval), VT(avfval), VT(av1fval), VT(av2fval), VT(initval))
		{
		}

	VtavfiltDYN(FT fval, FT avfval, FT av1fval, FT initval) :
		VtavfiltDYN(VT(fval), VT(avfval), VT(av1fval), VT(initval))
		{
		}

	VtavfiltDYN(FT fval, FT avfval, FT initval) :
		VtavfiltDYN(VT(fval), VT(avfval), VT(initval))
		{
		}

	VtavfiltDYN(const me& src) :
		curval(src.curval), curtarget(src.curtarget),
			filt(src.filt), avfilt(src.avfilt), av1filt(src.av1filt), av2filt(src.av2filt), av3filt(src.av3filt), av4filt(src.av4filt) {}

	~VtavfiltDYN() {}
	
	me&
	operator=(const me& src)
	{
		if (&src == this)
			return *this;

		curval = src.curval;
		curtarget = src.curtarget;
		filt = src.filt;
		avfilt = src.avfilt;
		av1filt = src.av1filt;
		av2filt = src.av2filt;
		av3filt = src.av3filt;
		av4filt = src.av4filt;
		return *this;
	}

	void
	newinit(bf fval, bf avfval, bf av1fval = bf(0), bf av2fval = bf(0), bf av3fval = bf(0), bf av4fval = bf(0))
	{
		filt = fval;
		avfilt = avfval;
		av1filt = av1fval;
		av2filt = av2fval;
		av3filt = av3fval;
		av4filt = av4fval;
	}

	void
	setfilt(bf f)
	{
		filt = f;
	}

	void
	setfilt(FT f)
	{
		filt = bf(f);
	}

	void
	setavfilt(bf f)
	{
		avfilt = f;
	}

	void
	setav1filt(bf f)
	{
		av1filt = bf(f);
	}

	void
	setav2filt(bf f)
	{
		av2filt = bf(f);
	}

	void
	setav3filt(bf f)
	{
		av3filt = bf(f);
	}

	void
	setav4filt(bf f)
	{
		av4filt = bf(f);
	}

	void
	setavfilt(FT f)
	{
		avfilt = bf(f);
	}

	void
	setav1filt(FT f)
	{
		av1filt = bf(f);
	}

	void
	setav2filt(FT f)
	{
		av2filt = bf(f);
	}

	void
	setav3filt(FT f)
	{
		av3filt = bf(f);
	}

	void
	setav4filt(FT f)
	{
		av4filt = bf(f);
	}

	bf
	getfilt(void) const
	{
		return filt;
	}

	bf
	getavfilt(void) const
	{
		return avfilt;
	}

	bf
	getav1filt(void) const
	{
		return av1filt;
	}

	bf
	getav2filt(void) const
	{
		return av2filt;
	}

	bf
	getav3filt(void) const
	{
		return av3filt;
	}

	bf
	getav4filt(void) const
	{
		return av4filt;
	}

	void
	setfilt(bf dcy, bf atk0, bf atk1 = bf(FT(0)), bf atk2 = bf(FT(0)), bf atk3 = bf(FT(0)), bf atk4 = bf(FT(0)))
	{
		filt = dcy;
		avfilt = atk0;
		av1filt = atk1;
		av2filt = atk2;
		av3filt = atk3;
		av4filt = atk4;
	}

	void
	setfilt(FT dcy, FT atk0, FT atk1 = FT(0), FT atk2 = FT(0), FT atk3 = FT(0), FT atk4 = FT(0))
	{
		filt = bf(dcy);
		avfilt = bf(atk0);
		av1filt = bf(atk1);
		av2filt = bf(atk2);
		av3filt = bf(atk3);
		av4filt = bf(atk4);
	}

	static bi optdir(bf newval, bf orig)
	{
		if (NEG)
			return Vlt(newval, orig);
		else	
			return Vgt(newval, orig);
	}

	bf ALWAYS_INLINE
	next(const bf& v, bool opt = false)
	{

		bf rtval;
		bf tgdiff = v - curtarget;
		curtarget += tgdiff * filt;
		bf vdiff = v - curval;
		bf xtragain = opt ? tgdiff : vdiff;
		bf vfpos = Vselect(Vgt(xtragain, bf(0.0)), xtragain, bf(-1.0f) * xtragain);
		//bf vfpos = xtragain;
		bf TMPavfilt = avfilt +
			Vmax(vfpos * (av1filt + (vfpos * (av2filt + (vfpos * av3filt + (vfpos * av4filt))))), bf(0.0f));
		bf infilt{};

		if (NEG == true)
		{
			bf atkflags = Vlt(vdiff, bf(FT(0.0)));
		}
		else if (NEG == false)
		{
			bf atkflags = Vgt(vdiff, bf(FT(0.0)));
			infilt = Vselect(atkflags, TMPavfilt, filt);
		}
		else
		{
			infilt = TMPavfilt + (opt ? bf(0.0) : filt);
		}

		bf cchange = infilt * vdiff;
		bf newvalnoclip = curval + cchange;

		bf pattack = Vgt(cchange, bf(FT(0.0)));
		bf nattack = Vlt(cchange, bf(FT(0.0)));

		rtval = Vselect(pattack, Vmin(v, newvalnoclip), newvalnoclip);
		rtval = Vselect(nattack, Vmax(v, newvalnoclip), newvalnoclip);

		curval = rtval;
		return rtval;
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

	template <typename PV>
	PV operator()(const PV& input, int opt)
	{
		return next(input, opt);
	}

	bf
	getcurval(void) const
	{
		return curval;
	}

	void 
	setcurval(bf in = bf(FT(0.0)))
	{
		curval = in;
	}

	void 
	setcurval(FT in = FT(0.0))
	{
		curval = bf(in);
	}
};
#endif

template <typename VDEF, int NEG = false>
class Vtavfilt GCQ
{
	typedef varrayT<VDEF> VT;
	typedef IvarrayT<VDEF> IVT;
	typedef typename VDEF::FT FT;
	typedef typename VDEF::IT IT;

	using VAFT=VT;
	using VAIT=IVT;

	VAFT curval;
	VAFT filt;
	VAFT avfilt;

	typedef Vtavfilt<VDEF,NEG> me;

public:

	using bf=VAFT;
	using bi=VAIT;
	static constexpr int VS = VDEF::VS;

	Vtavfilt() : curval(), filt(), avfilt() {}

	Vtavfilt(bf fval, bf avfval, bf initval) :
		curval(initval), filt(fval), avfilt(avfval)
		{
		}

	Vtavfilt(FT fval, FT avfval, FT initval) :
		Vtavfilt(VT(fval), VT(avfval), VT(initval))
		{
		}

	Vtavfilt(const me& src) :
		curval(src.curval), filt(src.filt), avfilt(src.avfilt) {}

	~Vtavfilt() {}
	
	me&
	operator=(const me& src)
	{
		if (&src == this)
			return *this;

		curval = src.curval;
		filt = src.filt;
		avfilt = src.avfilt;
		return *this;
	}

	void
	newinit(bf fval, bf avfval)
	{
		filt = fval;
		avfilt = avfval;
	}

	void
	setfilt(bf f)
	{
		filt = f;
	}

	void
	setfilt(FT f)
	{
		filt = bf(f);
	}

	void
	setavfilt(bf f)
	{
		avfilt = f;
	}

	void
	setavfilt(FT f)
	{
		avfilt = bf(f);
	}

	bf
	getfilt(void) const
	{
		return filt;
	}

	void
	setfilt(bf dcy, bf atk0)
	{
		filt = dcy;
		avfilt = atk0;
	}

	void
	setfilt(FT dcy, FT atk0)
	{
		filt = bf(dcy);
		avfilt = bf(atk0);
	}

	static bi optdir(bf newval, bf orig)
	{
		if (NEG)
			return Vlt(newval, orig);
		else	
			return Vgt(newval, orig);
	}

	bf 
	next(const bf& v)
	{

		bf vdiff = v - curval;
		bf infilt{};

		if (NEG == true)
		{
			bf atkflags = Vlt(vdiff, bf(FT(0.0)));
			infilt = Vselect(atkflags, avfilt, filt);
		}
		else if (NEG == false)
		{
			bf atkflags = Vgt(vdiff, bf(FT(0.0)));
			infilt = Vselect(atkflags, avfilt, filt);
		}

		bf cchange = infilt * vdiff;
		bf rtval = curval + cchange;

		curval = rtval;
		return rtval;
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

	bf
	getcurval(void) const
	{
		return curval;
	}

	void 
	setcurval(bf in = bf(FT(0.0)))
	{
		curval = in;
	}

	void 
	setcurval(FT in = FT(0.0))
	{
		curval = bf(in);
	}
};


#undef NOINLINE

//
// If we are running a recent GCC, IF we want 'noinline', we probably
// want 'noclone' also.
//
#if defined(__GNUC__) && (__GNUC__ > 5)
#define NOINLINE __attribute__((noinline,noclone))
#else
#define NOINLINE __attribute__((noinline))
#endif


//
// default initial linear input rms value
//
static constexpr float defaultrms = 1.0f;

//
// default initial dB input value (relative to 1 as 0dB)
//
static constexpr float defaultdB = 0.0f;

//
// The Vector/float types used to be selectable due to problems with the control
// signal filter dynamic ranges.  That problem has been worked around, so we
// use a fixed set of types.  Now, we use the varray series of types, and
// can easily select the types as needed.  Also, dynamic range issues are in the past.
//

//
// Maximum number of espansion steps -- only one is required for kind-of DolbyA decoding,
// since bugfixes make the compression/expansion
// work correctly!!!
//
	static constexpr int MAXEXP = 16;

static Asample
ninetymatrix(const Asample& in, bool mode = false)
{
					if (mode == false)
					{
	               float mid = (in.item0() + in.item1());
  		             float side = (in.item0() - in.item1());

						mid *= 0.281f;
						side *= 0.959f;

						return Asample(mid + side, mid - side) * 0.707f * 1.02f;
					}
					else
					{
						Asample rtv(2.302f*in.item0() + 1.256f*in.item1(), 1.256f * in.item0() + 2.302f*in.item1());
						return rtv * 1.25 * 0.707f;
					}
};

static Asample
msmatrix(const Asample& in, bool mode = false)
{
	
	if (mode == false)
	{
		float mid = (in.item0() + in.item1());
		float side = (in.item0() - in.item1());

		return Asample(mid * 0.707f, side * 0.707f);
	}
	else
	{
		float left = (in.item0() + in.item1());
		float right = (in.item0() - in.item1());
		return Asample (left * 0.707f, right * 0.707f);
	}
}


//
// Manifest constants for configuration expressions handled by the procexpcmd procedure/expcfgT class
//

	static constexpr int DMODE_SKIP = 0;
	static constexpr int DMODE_LR = 1;
	static constexpr int DMODE_90 = 2;
	static constexpr int DMODE_MS = 3;
	static constexpr int DMODE_LRGAINONLY = 4;
	static constexpr int DMODE_90GAINONLY = 5;
	static constexpr int DMODE_MSGAINONLY = 6;
	static constexpr int DMODE_LRWONLY = 7;
	static constexpr int DMODE_90WONLY = 8;
	static constexpr int DMODE_MSWONLY = 9;
	static constexpr int DMODE_X = 10;

	static constexpr const char * const modeid[12] {
		"SKIP", "LEFT-RIGHT", "IN-BETWEEN", "MID-SIDE", "LR-GAINONLY", "BTW-GAINONLY", "MS-GAINONLY", "LR-WIDTHONLY", "BTW-WIDTHONLY", "MS-WIDTHONLY", "XALL", "NONE"
	};


//
// class/data structure that controls the decoding (expanding).  One of these per step
//
	struct expcfgT
	{
		int dmode;
		float dBthresh;
		float LINthresh;
		float dBgainbefore;
		float LINgainbefore;

		expcfgT() : dmode(DMODE_SKIP), dBthresh(0.0f), LINthresh(1.0f), dBgainbefore(0.0f), LINgainbefore(1.0f) {}

		void setgaindB(float dBgain)
		{
			dBgainbefore = dBgain;
			LINgainbefore = dB2lin(dBgain);
		}

		void setgainLIN(float LINgain)
		{
			LINgainbefore = LINgain;
			dBgainbefore = lin2dB(LINgain);
		}

		void setthreshdB(float dBval)
		{
			dBthresh = dBval;
			LINthresh = dB2lin(dBval);
		}

		void setthreshLIN(float LINval)
		{
			LINthresh = LINval;
			dBthresh = lin2dB(LINval);
		}

		void dump(void) const
		{
			fprintf(stderr,"%10.10s dBthresh: %6.2f, dBgain: %6.2f\n", modeid[dmode], dBthresh, dBgainbefore);
		}
	};

//
// Number of expansion steps
//
	int numexp = 0;

//
// Array containing the steps
//
	expcfgT expcfg[MAXEXP];


//
// dump all of the dolby steps
//
	void 
	dumpallexp(void)
	{
		for (int i = 0; i < numexp; i++)
		{
			fprintf(stderr,"step%d: ", i);
			expcfg[i].dump();
		}
	}

//
// Return whether or not this is a dolby processor or if it is a fixed gain
//
	bool
	isdexpand(int index)
	{
		int dm = expcfg[index].dmode;
		if (dm == DMODE_LR || dm == DMODE_90 || dm == DMODE_MS || dm == DMODE_X)
			return true;
		return false;
	}

	bool
	isxexpand(int index)
	{
		int dm = expcfg[index].dmode;
		if (dm == DMODE_X)
			return true;
		return false;
	}

//
// parses the expansion control commands -- a string array of
// (<cmd>,<thresh>,<gain>), producing the expcfg array, with a count
// of numexp.
//
	int 
	procexpcmd(const char *incmd)
	{
		enum expcmdst {NOST = 0, INCMD = 1, INTHRESH = 3, INGAIN = 5, FINISH=999 };

		int incp = 0;
		char inc;
		expcmdst dcst{NOST};

		static constexpr int MAXTMPBUFSIZE = 128;
		char tmpbuf[MAXTMPBUFSIZE];

		int tmpbufcnt = 0;
		expcfgT* curexpcfg = &expcfg[0];
		numexp = 0;
		
		//
		// trivial parsing states
		//

nextitem:
		while ((dcst != FINISH) && (inc = incmd[incp]) != '\0')
		{
			incp++;
			inc = tolower(inc);

			switch (dcst)
			{
	case NOST:
				if (isblank(inc))
					break;

				if (inc == '(')
				{
					curexpcfg = &expcfg[numexp];

					if (numexp == 0)
						*curexpcfg = expcfgT{};
					else
						*curexpcfg = expcfg[numexp-1];

					numexp++;

					dcst = INCMD;
					tmpbufcnt = 0;
				}
				break;

	case INCMD:
				if (isblank(inc))
					break;

				if (inc == ',' || inc == ')')
				{
					tmpbuf[tmpbufcnt] = '\0';
					if (tmpbufcnt != 0)
					{
						bool gainmode = false;
						bool widthmode = false;
						int tptr = 0;

						char tcmd = tmpbuf[tptr];

						if (tcmd == 'g')
						{
								gainmode = true;
								tptr++;
								tcmd = tmpbuf[tptr];
								if (tcmd == '\0')
									tcmd = 'l';
						}

						if (tcmd == 'w')
						{
								widthmode = true;
								tptr++;
								tcmd = tmpbuf[tptr];
								if (tcmd == '\0')
									tcmd = 'l';
						}

							
						if ((tcmd == 'l') || (tcmd == 'r'))
						{
							curexpcfg->dmode = widthmode ? DMODE_LRWONLY : (gainmode ? DMODE_LRGAINONLY : DMODE_LR);
						}
						else if ((tcmd == 'm') || (tcmd == 's'))
						{
							curexpcfg->dmode = widthmode ? DMODE_MSWONLY : (gainmode ? DMODE_MSGAINONLY : DMODE_MS);
						}
						else if ((tcmd == 'b') || (tcmd == 'i') || (tcmd == '9'))
						{
							curexpcfg->dmode = widthmode ? DMODE_90WONLY : (gainmode ? DMODE_90GAINONLY : DMODE_90);
						}
						else if (tcmd == 'x')
						{
							curexpcfg[0].dmode = DMODE_X;
							curexpcfg[1].dmode = DMODE_SKIP;
							curexpcfg[2].dmode = DMODE_SKIP;
							numexp += 2; /* X does three at a time, skip next two */
						}
					}

					tmpbufcnt = 0;

					if (inc == ')')
						dcst = NOST;
					else
						dcst = INTHRESH;
			
					break;
				}

				tmpbuf[tmpbufcnt] = inc;
				tmpbufcnt++;
				break;

	case INTHRESH:
				if (isblank(inc))
					break;

				if (inc == ',' || inc == ')')
				{
					tmpbuf[tmpbufcnt] = '\0';
					if (tmpbufcnt != 0)
					{
						curexpcfg->setthreshdB(atof(&tmpbuf[0]));
					}

					tmpbufcnt = 0;

					if (inc == ')')
						dcst = NOST;
					else
						dcst = INGAIN;

					break;
				}

				tmpbuf[tmpbufcnt] = inc;
				tmpbufcnt++;
				break;

	case INGAIN:
				if (isblank(inc))
					break;

				if (inc == ',' || inc == ')')
				{
					tmpbuf[tmpbufcnt] = '\0';
					if (tmpbufcnt != 0)
					{
						curexpcfg->setgaindB(atof(&tmpbuf[0]));
					}
					tmpbufcnt = 0;

					dcst = NOST;

					if ( inc == ',')
					{
						fprintf(stderr,"comma not expected after the gain parameter at character location %d\n", incp);
					}
					break;
				}
				tmpbuf[tmpbufcnt] = inc;
				tmpbufcnt++;
				break;

	case FINISH:
	default:
				break;
			}
			
		}
		return 0;
	}
		
//
// process and dump the commands
//
void
pdcmd(const char *incmd)
{
	procexpcmd(incmd);
	if (ac.info > 2) dumpallexp();
}


//
// Define the data types.   The vector mechanism is supported by the file varray.h -- and provides
// the various operations/conversion between all kinds of floatingpoint and integer SIMD types
//
#if defined(EIGHTWIDE)
using VDEF4 = TYPEvd4x1;			// this is the descriptor for the vector type
using VFP4 = VDEF4::FT;				// this is the underlying floating point type in the vector
using vtype4 = varrayT<VDEF4>;	// the actual vector type that we are using

using VDEF8 = TYPEvf8x1;			// this is the descriptor for the vector type
using VFP8 = VDEF8::FT;				// this is the underlying floating point type in the vector
using vtype8 = varrayT<VDEF8>;	// the actual vector type that we are using

#else

using VDEF4 = TYPEvf4x1;			// this is the descriptor for the vector type
using VFP4 = VDEF4::FT;				// this is the underlying floating point type in the vector
using vtype4 = varrayT<VDEF4>;	// the actual vector type that we are using
#endif





namespace da
{
	//
	// basic attack/decay times derived from lots of listening of apparently appropriately
	// 'compressed' material.
	// the approximate times are 1/(value * 44100)
	//
	static constexpr VFP4 satk = 0.0102;			// LF/MF decay time for fast attacks
	static constexpr VFP4 sdcy = 0.00038;		// LF/MF attack/decay with nonlinear components below
	static constexpr VFP4 fatk = 0.0202;			// HF decay time for fast attacks
	static constexpr VFP4 fdcy = 0.00077;		// HF attack/decay with nonlinear components below


	static constexpr VFP4 hpfsdcy0 = sdcy * 0.34;
	static constexpr VFP4 hpffdcy0 = fdcy * 0.34;

	static constexpr VFP4 hpfrcomp = 0.64;
	//	static constexpr VFP4 hpfrcomp = 0.67;
	static constexpr VFP4 hpfsdcy = sdcy * hpfrcomp;
	static constexpr VFP4 hpffdcy = fdcy * hpfrcomp;

	using TfltSA = Vtavfilt<VDEF4>;	// this is a simple single TC attack/decay with a nonlinear attack

#include "recursexp.h"

//
// Expander routine
//
// IN THIS FILE -- ALL FILTER COEFFICIENTS ARE FOR 44.1K, even if the processor doesnt' support 44.1K.  Other
// portions of the code do the conversion for the sample rates.
//
class dbexpand
{
	//
	//  Direct or pseduo feedback mode
	//
#define FBMODE 0
static constexpr int MAXGAINHPF=64;

	//
	// general types used for gains/levels for the 4 channels
	//
	// VDEF4 is the varrayT descriptor, while vtype4 (a single vector of 4 VFP4s)
	// is the actual type of the data
	//

	using GAINHPFTYPE = hpfT<vtype4, VFP4>;
	//
	// modified single attack/release control type -- with the nonlinear elements
	// added for either speeding up an overly soft attack, or slowing down an overly
	// quick attack.  This is obviously the scheme used by Dolby to help with
	// avoiding the strong attack glitches (a form of intermod.)
	//
	// I am echoing the type of attack/decay used originally only for the effect on
	// the sound, not to avoid glitches.  I have a proprietary scheme that avoids
	// the 'glitches' or 'crackles' WITHOUT modifying the attack times (my proprietary
	// scheme can work with any attack/decay times without glitching.)
	//

	//
	// gain control/level detect filters -- described below
	//
	TfltSA dfiltlLG;
	TfltSA dfiltrLG;
	TfltSA dfiltlLNB;
	TfltSA dfiltrLNB;
	TfltSA dfiltlA;
	TfltSA dfiltrA;

	TfltSA LRdfiltlLG;
	TfltSA LRdfiltrLG;
	TfltSA LRdfiltlLNB;
	TfltSA LRdfiltrLNB;
	TfltSA LRdfiltlA;
	TfltSA LRdfiltrA;

	TfltSA M90dfiltlLG;
	TfltSA M90dfiltrLG;
	TfltSA M90dfiltlLNB;
	TfltSA M90dfiltrLNB;
	TfltSA M90dfiltlA;
	TfltSA M90dfiltrA;

	TfltSA MSdfiltlLG;
	TfltSA MSdfiltrLG;
	TfltSA MSdfiltlLNB;
	TfltSA MSdfiltrLNB;
	TfltSA MSdfiltlA;
	TfltSA MSdfiltrA;

	//
	// threshold offset (lowest level threshold) LINear and dB
	//
	//static constexpr VFP4 defaultdBToffset = -41.5;
	static constexpr VFP4 defaultdBToffset = -41.0;

	VFP4 LINToffset, dBToffset;
	vtype4 EXpvtI;
	vtype4 dBEXpvtI;
	vtype4 INPthreshold; 
	vtype4 dBINPthreshold; 
	vtype4 XINPthreshold; 
	vtype4 XdBINPthreshold; 

	//
	// sample count
	//
	int scnt;

	//
	// high pass filters/DC&LF removal
	//
	hpf4As inputfilt;
	hpf4As outputfilt0;
	hpf4As outputfilt1;

	//
	// band splitter
	//
	dATsingle infilt;

	//
	// gain passed back
	//
	vtype4 pgainl, pgainr;
	vtype4 LRpgainl, LRpgainr;
	vtype4 M90pgainl, M90pgainr;
	vtype4 MSpgainl, MSpgainr;
	int curgainpf;

	//
	// previous sample
	//
	Asample prevsample;

	//
	// output cleanup filters (different curves based on sample rates 
	// mostly 'gone' by 30kHz at higher sample rates.)
	//
	lpf20500 ofiltA, ofiltB;

	//
	// index for gain/index for thresholds
	//
	int myindex, mytindex;
	//
	// Special modes
	//
	unsigned mode;

	V0recursefilt rfpl, rfpr;
	V0recursefilt LRrfpl, LRrfpr;
	V0recursefilt M90rfpl, M90rfpr;
	V0recursefilt MSrfpl, MSrfpr;

	//
	// nonlinear constants for the visible attack/decay times
	// painfully/iteratively derived.  Surprisingly, the third order
	// below is critical -- otherwise moderate attacks will seem
	// blunted.  Generally, the most critical number is the 2nd order,
	// and seems close to correct -- perhaps needs a little optimization.
	// At this point, the difference between o3 being 0.0 or the value specifed below
	// is more important than a small 5-10 difference in o2.
	//
	// BOTH atko2mult and atko3mult could use a bit more tuning.  They are definitely in the
	// 'ballpark.'
	//

	static constexpr VFP4 atko2mult = 0.00;
	static constexpr VFP4 atko3mult = 0.00;
	static constexpr VFP4 atko4mult = 0.00;
	static constexpr VFP4 atko5mult = 0.00;
	//
	// fast averaging -- really meant to help avoid the noise from
	// any inaccurate or other inconsistency in transcendental functions
	//
	//static constexpr VFP4 ainavg = 0.330;
	//
	// this soft attack was added directly to the closest gain control.  It was originally meant
	// to slightly soften the 'edge' on the sound without really affecting the transient
	// attacks.  Instead of running with 100usec attack time, I set up a 1msec attack time
	// that moves more quickly on a sharp attack on the compression side. Seems slghtly more
	// natural sound.  (I wonder if this helps to undo some of the FAST clipping?)  I'd think
	// that it would be the opposite...  Not sure, but it does seem to sound a LITTLE better
	// to me.
	//
	// NOW....  The situation has changed to where the up-front timeconstant really does help ALOT:
	//
	// Basically -- these tunables seem to help get rid of the harsh edge in some singing (EXISTING
	// BEFORE ATTEMPTING TO DECODE.)  Apparently, there is a disparity or phase issue in the encoding
	// (compression.)  When decoding, it appears that it sounds much smoother to retard (slightly)
	// the attack for the low/middle freqs, and only slightly filter the HF ranges.  This mod helps
	// produce a smoother/better sound -- MUCH MUCH smoother than the original non-decoded version.
	// IF this filtering is not used, then the decoded version sounds harsh like the non-decoded version.
	// For example, it is like with prominent vocals of Karen Carpenter, her voice has a 'saw'
	// on the edge.  (Harsh sound.)  This has been noticeable to me in recent years -- even though
	// my hearing is a little worse.  By decoding the vocals with these delay/attack parameters
	// directly on the gain control -- it appears that a big part of the harshness disappears.
	// I have heard this kind of effect on my compressors when the frequency ranges have inconsistent
	// characteristics.  This is why I tried this experiment.
	//
	// (THIS WILL LIKELY BENEFIT FROM MORE TUNING -- BUT I HAVE MORE TO DO -- IT GOT RID OF AT LEAST 75%
	// OF THE EXCESS EDGINESS.)
	//
	// -- it might be better to modify the main timeconstants above, but then the dynamics might be more
	// negatively affected.
	//

	//static constexpr VFP4 ainfst = 0.220;
	static constexpr VFP4 ainfstr = 0.50f;
	static constexpr VFP4 ainfst = 0.075;
	static constexpr VFP4 ainfsts = 0.025;
	static constexpr VFP4 ainfsta = 0.220;
	static constexpr VFP4 ainavg = 0.00330;
	static constexpr VFP4 ainavghf = 0.0250;
	static constexpr VFP4 aino2mult = 8.00;
	static constexpr VFP4 aino3mult = 0.50;
	static constexpr VFP4 aino4mult = 0.00;
	static constexpr VFP4 aino5mult = 0.50;


	// note that these TCs above are so fast that they don't have strong affect on the dynamics, but
	// can have significant impact on the sound smoothness.


//
// these simple inline functions with 'gf' in their names are the gainfunctions.  These
// are the dB/dB conversions that implement the expansion ratios.  This doesn't use linear
// dB/dB expansion, so these formulas do the calcuations.
//
// NOTE THAT THESE FORMULAS ARE EXPERIMENTS AND REQUIRED NUMEROUS LISTENING ITERERATIONS TO DERIVE
// THESE CURVES.   Painfully derived and historical code tests -- there is alot of chaff in these snippets,
// but I don't have the heart to remove them.
//
	template <typename DT>
	DT fullgf(DT lvl)
	{
		return (DT(-0.0) - (DT(0.046986) * lvl)) - (DT(0.006598) * lvl * lvl);
	}

	template <typename DT>
	DT fgfe(DT llvl)
	{
		DT gfval = fullgf(llvl);
		DT extgf = Vselect(Vgt(llvl, DT(-7.0)), DT(-0.50000) + (llvl - DT(-7.0)) * DT(0.100), gfval);
		extgf = Vselect(Vlt(llvl, DT(-40.0)), (DT(-9.17736) + (llvl + DT(40.0))* DT(1.0)), extgf);
		extgf = Vmin(extgf, DT(0.0));
		return extgf;
	}

	VFP4 fgfe(VFP4 llvl)
	{
		vtype4 rtval(fgfe(vtype4(llvl)));
		return rtval[0];
	}

public:
	dbexpand(int index = 2) /* NOINLINE */ :

//
// dfiltlLG/dfiltrLG impliment the quick attack/ msec release component of the expanders general attack/release scheme
//
// in these attack/release schemes, the following describe the actions:
//
// the first line is the release time (ignore the manifest symbols satk/fatk -- they are release
//                times when in the first line)
// the second line is the attack time -- 1.0 is infinitely fast
// the third line is the 2nd order nonlinear component of the attack time
// the 4th line is the 3rd order nonlinear component of the attack time
// if the 3rd or 4th lines are missing, then they are zero -- zero is effectively disabled.
//
//
// These are simple 1st order nonlinear (seperate attack/decay filters) with the following general concept
// for fixed attack/decay times.  The nonlinear versions are a little more complicated (but all encapsulated into
// a single class.)  The method is SOMETHING like this:
//
// if (attack)
//   outputval += (newval - oldoutputval) * attackcoeff;
// else if (decay)
//   outputval += (newval - oldoutputval) * decaycoeff;
//
// There are also some infinitely (almost truly) more complex filter classes available in my audio
// gain control filtering package for very good audio processing.  I am amazed that R Dolby is
// using such simple filters and getting such good results (but his nonlinear
// attack trick is very good to help avoid some of the worst intermod issues!!!!)
//
		dfiltlLG(
			vtype4({sratetmod(satk), sratetmod(satk), sratetmod(fatk), sratetmod(fatk)}),
			vtype4({sratetmod(satk), sratetmod(satk), sratetmod(fatk), sratetmod(fatk)}),
			vtype4(VFP4(0.1))),
		dfiltrLG(
			vtype4({sratetmod(satk), sratetmod(satk), sratetmod(fatk), sratetmod(fatk)}),
			vtype4({sratetmod(satk), sratetmod(satk), sratetmod(fatk), sratetmod(fatk)}),
			vtype4(VFP4(0.1))),

//
// dfiltlLNB/dfiltrLNB implement the more important dynamic attack/fixed release scheme
// (note the use of the 2nd order and 3rd order components
//
		dfiltlLNB(
			vtype4({sratetmod(sdcy), sratetmod(sdcy), sratetmod(fdcy), sratetmod(fdcy)}),
			vtype4({sratetmod(sdcy), sratetmod(sdcy), sratetmod(fdcy), sratetmod(fdcy)}),
			vtype4(0.1)),
		dfiltrLNB(
			vtype4({sratetmod(sdcy), sratetmod(sdcy), sratetmod(fdcy), sratetmod(fdcy)}),
			vtype4({sratetmod(sdcy), sratetmod(sdcy), sratetmod(fdcy), sratetmod(fdcy)}),
			vtype4(0.1)),


//
// this is the gain control filter -- might or might not be used depending upon listening tests
// dfiltlA is copied to all of the other filters to make the code less tedious.
//

		dfiltlA(
			vtype4({sratetmod(VFP4(ainfsts)), sratetmod(VFP4(ainfst)), sratetmod(VFP4(ainfst)), sratetmod(VFP4(ainfst))}),
			vtype4({sratetmod(VFP4(ainfsts)), sratetmod(VFP4(ainfst)), sratetmod(VFP4(ainfst)), sratetmod(VFP4(ainfst))}),
			vtype4(1.0)),

		dfiltrA(),

		LRdfiltlLG(),
		LRdfiltrLG(),
		M90dfiltlLG(),
		M90dfiltrLG(),
		MSdfiltlLG(),
		MSdfiltrLG(),

		LRdfiltlLNB(),
		LRdfiltrLNB(),
		M90dfiltlLNB(),
		M90dfiltrLNB(),
		MSdfiltlLNB(),
		MSdfiltrLNB(),

		LRdfiltlA(),
		LRdfiltrA(),
		M90dfiltlA(),
		M90dfiltrA(),
		MSdfiltlA(),
		MSdfiltrA(),

		LINToffset(),
		dBToffset(),

		EXpvtI(),
		dBEXpvtI(),
		dBINPthreshold(),
		INPthreshold(),
		XdBINPthreshold(),
		XINPthreshold(),


//
// this is the LF filter removing DC (one for each freq band)  The values are the LF filter coefficients
//
		inputfilt(sratetmod(0.000225), sratetmod(0.00045), sratetmod(0.010), sratetmod(0.010)),
		//outputfilt0(sratetmod(0.00030), sratetmod(0.00080), sratetmod(0.045), sratetmod(0.10)),
		//outputfilt1(sratetmod(0.00030), sratetmod(0.00080), sratetmod(0.045), sratetmod(0.10)),
		outputfilt0(sratetmod(0.00038), sratetmod(0.00038), sratetmod(0.00038), sratetmod(0.00038)),
		outputfilt1(sratetmod(0.00038), sratetmod(0.00038), sratetmod(0.00038), sratetmod(0.00038)),
//
// this is the input band splitting filter
//
		infilt(),

//
// sample count
//
		scnt(0),

//
// gain passed back from previous sample
//
		pgainl(1.0), pgainr(1.0),
		LRpgainl(1.0), LRpgainr(1.0),
		M90pgainl(1.0), M90pgainr(1.0),
		MSpgainl(1.0), MSpgainr(1.0),
		curgainpf(0),

		ofiltA(), ofiltB(),
		myindex(index), mytindex(index),
		mode(0U),
		rfpl(), rfpr(),
		LRrfpl(), LRrfpr(),
		M90rfpl(), M90rfpr(),
		MSrfpl(), MSrfpr()

	{

			{
				char *cval = getenv("DBAMODE");
				if (cval != nullptr)
				{
					mode = unsigned(strtol(cval, nullptr, 16));
//					if (ac.info) fprintf(stderr,"DBAMODE: 0x%0x\n", mode);
				}
			}

			if (mode & 0x4)
			{
				outputfilt0.setcoeffs(sratetmod(0.00038), sratetmod(0.00060), sratetmod(0.033), sratetmod(0.1));
				outputfilt1.setcoeffs(sratetmod(0.00038), sratetmod(0.00060), sratetmod(0.033), sratetmod(0.1));
			}
			else if (mode & 0x8)
			{
				outputfilt0.setcoeffs(sratetmod(0.00022), sratetmod(0.00038), sratetmod(0.025), sratetmod(0.050));
				outputfilt1.setcoeffs(sratetmod(0.00022), sratetmod(0.00038), sratetmod(0.025), sratetmod(0.050));
			}

			dBToffset = defaultdBToffset;
			LINToffset = dB2lin(dBToffset);

			EXpvtI = vtype4(expcfg[mytindex].LINthresh);
			dBEXpvtI = vtype4(expcfg[mytindex].dBthresh);

			dBINPthreshold = dBEXpvtI + vtype4(dBToffset);
			INPthreshold = dB2lin(dBINPthreshold);

			dfiltrLG = dfiltlLG;
			dfiltrLNB = dfiltlLNB;
			dfiltrA = dfiltlA;

			LRdfiltlLG = dfiltlLG;
			LRdfiltrLG = dfiltrLG;
			LRdfiltlLNB = dfiltlLNB;
			LRdfiltrLNB = dfiltrLNB;
			LRdfiltlA = dfiltlA;
			LRdfiltrA = dfiltrA;

			M90dfiltlLG = dfiltlLG;
			M90dfiltrLG = dfiltrLG;
			M90dfiltlLNB = dfiltlLNB;
			M90dfiltrLNB = dfiltrLNB;
			M90dfiltlA = dfiltlA;
			M90dfiltrA = dfiltrA;

			MSdfiltlLG = dfiltlLG;
			MSdfiltrLG = dfiltrLG;
			MSdfiltlLNB = dfiltlLNB;
			MSdfiltrLNB = dfiltrLNB;
			MSdfiltlA = dfiltlA;
			MSdfiltrA = dfiltrA;
	}

	struct vtype4pair
	{
		vtype4 v0, v1;
	};

	vtype4pair
	getToffset(float dBTin)
	{
			EXpvtI = vtype4(expcfg[mytindex].LINthresh);
			dBEXpvtI = vtype4(expcfg[mytindex].dBthresh);

			LINToffset = dB2lin(dBTin);
			dBToffset = dBTin;
			vtype4 dBthresh = dBEXpvtI + vtype4({dBToffset, dBToffset, dBToffset - VFP4(10.0), dBToffset - VFP4(10.0)});
			//vtype4 dBthresh = dBEXpvtI + vtype4({dBToffset, dBToffset, dBToffset, dBToffset});

			return vtype4pair {
				dBthresh, dB2lin(dBthresh)
			};
	}

	void 
	setToffset(float dBTin)
	{
			vtype4pair v = getToffset(dBTin);

			dBINPthreshold = v.v0;
			INPthreshold = v.v1;

			v = getToffset(dBTin - 10.0);
			//v = getToffset(dBTin);
			XdBINPthreshold = v.v0;
			XINPthreshold = v.v1;
	}

#if 0
	void 
	setToffset(float dBTin)
	{
			EXpvtI = vtype4(expcfg[mytindex].LINthresh);
			dBEXpvtI = vtype4(expcfg[mytindex].dBthresh);

			LINToffset = dB2lin(dBTin);
			dBToffset = dBTin;

			dBINPthreshold = dBEXpvtI + vtype4({dBToffset, dBToffset, dBToffset - VFP4(10.0), dBToffset - VFP4(10.0)});
			INPthreshold = dB2lin(dBINPthreshold);
	}
#endif

	void
	setindex(int idx)
	{
		myindex = idx;
	}

	void
	settindex(int idx)
	{
		mytindex = idx;
		setToffset(dBToffset);
	}

	static constexpr VFP4 mingain = 0.100;  // this is the number that programs the maximum loss (expansion)

//
// An enterprising person might be able to put this into a simple plugin.   There is alot of code
// that isn't presented here (but is in supporting routines) like audio filters, the actual code
// for attack/decay filters, and the various tricky parts like the gain management code that
// avoids most of the AGC distortions.  I didn't enable the anti-ripple code because I think
// that the compliment of this expander expects that the ripple is present.  However, the
// other kinds of distortion continue to be suppressed (or avoided.)
//
	vtype4
	nextlevel(const Asample& siginA, Asample& sigout)
	{
			Asample sigin[4];

			//
			// Take the simple stereo input signal, and split into 4 freq bands as represented
			// by 'sigin'.
			//
			infilt(siginA, sigin);
			inputfilt(sigin, sigin);

			//
			// This is the output signal that is built from the 4 output channels.
			//
			Asample sigouttmp{};

			//
			// This is the 4 channel representation of the output signal
			//
			Asample sigoutbuf[4];

			//
			// ofiltA/ofiltB do some removal of any resulting out of band stuff.
			// oddly -- the filters are actually much more effective near the
			// highest audio frequencies at the lower sample rates.  However,
			// I suspect that the decoder otherwise generally works best at 96k.
			// 192k requires twice the CPU, and I doubt that there is not
			// much benefit.
			// (at 44.1k samples/sec, the filter falls to -60dB by 21000,
			// while it sits fat and happy at about -35dB at 30kHz with
			// a 96k sample rate -- odd, for otherwise the same filter
			// parameters, heh?  Thats the world of digital filters.)
			//
			// I could have done a really good brickwall at 20500Hz, and just
			// might do so in the future -- but it does require lots of taps,
			// and might slow down the program a bit when running at higher
			// sample rates.  As is, the two filters are only 4th order IIR,
			// elliptical with a very little bit of ripple.  Since there are
			// two filters, I guess that the -35dB at 30kHz becomes -70dB,
			// and the middling frequencies of 25kHz are pretty well suppressed.
			// That is why I did two filters instead of one.  And also, back
			// when I thought that we needed to do the multi-axis version of
			// the expander, I didn't want subsequent steps to be interfered
			// with by the earlier steps.
			//
			// Also, and most importantly -- earlier versions of the code used
			// more real feedback of the actual audio -- and could fall into
			// oscillation if I made a bug, or even if the math on a processor
			// or library might work a little differently.  So, with so very
			// much attenuation at frequencies higher than might be noticed, it
			// might save a tweeter or some other physical device if the program
			// goes nuts.  It is very, very stable now.  The only thing that
			// could POSSIBLY go unstable is the gain control, not the audio
			// itself.
			//

			//
			// Here, we calculate the raw unfiltered signal squared
			// this is mostly just being used for absolute value
			// but in the past, this section of code came from a very
			// nice compressor which used carefully designed RMS
			// measurements.  Most code long gone -- I just do it
			// this way now.
			//
			vtype4 LINrmsinsql(VFP4(1.0));
			vtype4 LINrmsinsqr(VFP4(1.0));

			for (int i = 0; i < 4; i++)
			{
				Asample& stmp = sigin[i];

				float lsig = ((stmp.item0() * stmp.item0()));
				LINrmsinsql[i] = lsig;
				float rsig = ((stmp.item1() * stmp.item1()));
				LINrmsinsqr[i] = rsig;
			}

			//
			// All filters are linear (NOT SQUARED -- unlike most of my tech, where I tend to work
			// in the squared and dB domain.)
			//

			//
			// get the linear values
			//
			vtype4 LINrmsinl = vsqrt(LINrmsinsql);
			vtype4 LINrmsinr = vsqrt(LINrmsinsqr);

			//
			// add in the threshold as determined above
			//
			// we can choose from an additive threshold or a hard
			// threshold
			//
			LINrmsinl += INPthreshold;
			LINrmsinr += INPthreshold;

			//
			// Do the attack/release -- the LG filter does a fast attack, and
			// millisecond decay. The LNB filter does the slow attack/slow decay
			// These filters are put together in a contorted fashion.   The reason why
			// these filters are put together this way is based upon the DolbyA behavior.
			// If you change the code between 'DOLBYA FILTER CODE BEGIN/END', the
			// characteristics will not likely match a real DolbyA unit.  I don't normally
			// write my attack/decay filters like this, so there is a real reason for this
			// structure.
			//

// APPROX DOLBYA FILTER TIMING CODE BEGIN
			vtype4 fatmp{};
			vtype4 fbtmpl{};
			vtype4 fbtmpr{};

			fbtmpl = dfiltlLG(LINrmsinl);
			fatmp = Vselect(Vgt(LINrmsinl, fbtmpl), (LINrmsinl + fbtmpl) * VFP4(0.50), fbtmpl);
			vtype4 LINrmsrawtmpl = dfiltlLNB(fatmp);
			LINrmsrawtmpl = Vselect(Vgt(fatmp, LINrmsrawtmpl), (fatmp + LINrmsrawtmpl) * VFP4(0.50), LINrmsrawtmpl);

			fbtmpr = dfiltrLG(LINrmsinr);
			fatmp = Vselect(Vgt(LINrmsinr, fbtmpr), (LINrmsinr + fbtmpr) * VFP4(0.50), fbtmpr);
			vtype4 LINrmsrawtmpr = dfiltrLNB(fatmp);
			LINrmsrawtmpr = Vselect(Vgt(fatmp, LINrmsrawtmpr), (fatmp + LINrmsrawtmpr) * VFP4(0.50), LINrmsrawtmpr);
// APPROX DOLBYA FILTER TIMING CODE END

			static constexpr double ssoffset = 0.0;

			LINrmsrawtmpl = Vmax(rfpl.next(fbtmpl), LINrmsrawtmpl * vtype4(ssoffset));
			LINrmsrawtmpr = Vmax(rfpr.next(fbtmpr), LINrmsrawtmpr * vtype4(ssoffset));

			//
			// get the dB values for the dB calculations -- I derived my expansion curves
			// in the dB domain and haven't done the conversion to linear yet.
			// This is a technically unnecessary waste, but not too bad on recent machines.
			//
			// I have decided to remove the special threshold code (and support for the
			// differing LF/HF compression/expansion curve versions)
			//
			// All we need to do now is to convert the dB/dB gain curve to the
			// linear world.   Curve-fitting time again!!!  I dont' think that it is
			// possible to do the change algebraically (I already tried doing it myself
			// and using Maxima.)  I want to be fairly faithful to the curve -- it took
			// significant amount of time to tweak the curve until it sounded right.  If
			// I let myself, I could probably tweak it until the end of time.
			//
			// removing the lin-to-dB, applying the curve, then dB-to-lin, and just
			// doing a curve should greatly simplify the code and really speed it up.
			//
			vtype4 dBlvll = lin2dB(LINrmsrawtmpl);
			vtype4 dBlvlr = lin2dB(LINrmsrawtmpr);

			//
			// set the level to pivot against
			//
			dBlvll -= dBEXpvtI;
			dBlvlr -= dBEXpvtI;

			//
			// Here, we apply the compression ratio (currently in dB terms)
			//
			vtype4 dBnll = fgfe(dBlvll);
			vtype4 dBnlr = fgfe(dBlvlr);

			
			dBnll = Vmin(dBnll, vtype4(-0.50));
			dBnlr = Vmin(dBnlr, vtype4(-0.50));

			//
			// Convert back to linear
			//
			pgainl = dB2lin(dBnll);
			pgainr = dB2lin(dBnlr);

			//
			// Here we move the level values around for various legacy reasons.  If you are trying to figure out
			// WHY I am moving the values around as much as I am -- don't bother.  I am just doing it because the
			// old code had reasons to do so.  Now, there is a serious need for a code cleanup!!!
			//
			vtype4 dgainl = pgainl;
			vtype4 dgainr = pgainr;

			//
			// Slow down the gain changes a little (help to avoid aliasing) -- also avoid any errors
			// in secret sauce.
			//
			//pgainl = dfiltlA(pgainl);
			//pgainr = dfiltrA(pgainr);

			//
			// Here, we are calculating the gain for 9k-20k
			//
			VFP4 pgainl3 = (pgainl[3] + pgainl[2]) * VFP4(0.562);
			VFP4 pgainr3 = (pgainr[3] + pgainr[2]) * VFP4(0.562);

			//
			// Slow down the gain changes a little (help to avoid aliasing) -- also avoid any errors
			// in secret sauce.
			//
			pgainl = dfiltlA(pgainl);
			pgainr = dfiltrA(pgainr);

			//
			// Set the 10dB loss limit for the lower three bands
			//
			for (int i = 0; i < 3; i++)
			{
				pgainl[i] = Vmax(pgainl[i], VFP4(0.316));
				pgainr[i] = Vmax(pgainr[i], VFP4(0.316));
			}

			//
			// set the 15dB loss limit for the highest band
			//
			pgainl[3] = Vmax(pgainl3, VFP4(0.316 * 0.562));
			pgainr[3] = Vmax(pgainr3, VFP4(0.316 * 0.562));


			//
			// gain never higher than 0dB
			//
			pgainl = Vmin(pgainl, vtype4(VFP4(1.0)));
			pgainr = Vmin(pgainr, vtype4(VFP4(1.0)));
					

			//
			// seperate 3-9k and 9-20k
			//
			Asample sigtmp2 = sigin[2] - sigin[3];

			//
			// calculate new signal levels for each of the 4 bands
			// 0-80, 80-3k, 3k-9k, 9k-20k
			//
			if ((mode & 0x1) == 0)
			{
				sigoutbuf[0] = Asample{ sigin[0].item0() * float(pgainl[0]), sigin[0].item1() * float(pgainr[0]) };
				sigoutbuf[1] = Asample{ sigin[1].item0() * float(pgainl[1]), sigin[1].item1() * float(pgainr[1]) };
			}
			else
			{
				sigoutbuf[0] = sigin[0];
				pgainl[0] = 1.0f;
				pgainr[0] = 1.0f;
				sigoutbuf[1] = sigin[1];
				pgainl[1] = 1.0f;
				pgainr[1] = 1.0f;
			}
			sigoutbuf[2] = Asample{ sigtmp2.item0() * float(pgainl[2]), sigtmp2.item1() * float(pgainr[2]) };
			sigoutbuf[3] = Asample{ sigin[3].item0() * float(pgainl[3]), sigin[3].item1() * float(pgainr[3])};

			Asample ob[4];

			if (( mode & 0x2) == 0)
			{
				outputfilt0(sigoutbuf, ob);
				outputfilt1(ob, ob);
			}
			else
			{
				ob[0] = sigoutbuf[0];
				ob[1] = sigoutbuf[1];
				ob[2] = sigoutbuf[2];
				ob[3] = sigoutbuf[3];
			}


			//
			// build the output -- summing each of the 4 bands
			//
			if (mode & 0x100)
			{
				for (int i = 1; i < 4; i++)
				{
					sigouttmp += ob[i];
				}
			}
			else if (mode & 0x200)
			{
				for (int i = 0; i < 1; i++)
				{
					sigouttmp += ob[i];
				}
			}
			else
			{
				for (int i = 0; i < 4; i++)
				{
					sigouttmp += ob[i];
				}
			}

			sigout = ofiltA(ofiltB(sigouttmp));

			prevsample = sigouttmp;

			vtype4 avggain = (pgainl + pgainr) * vtype4(VFP4(0.50));
			return avggain;
	}

	struct Xnextlevelrtval
	{
			vtype4 v0, v1, v2;
	};

	Xnextlevelrtval
	Xnextlevel(const Asample& siginA, Asample& sigout)
	{
			Asample sigin[4];
			Asample siginLR[4], sigin90[4], siginMS[4];
			Asample sigoutLR[4], sigout90[4], sigoutMS[4];

			//
			// Take the simple stereo input signal, and split into 4 freq bands as represented
			// by 'sigin'.
			//
			infilt(siginA, sigin);

			inputfilt(sigin, sigin);

			for (int i = 0; i < 4; i++)
			{
				siginLR[i] = sigin[i];
				sigin90[i] = ninetymatrix(sigin[i], true);
				siginMS[i] = msmatrix(sigin[i], true);
			}

			//
			// This is the output signal that is built from the 4 output channels.
			//
			Asample sigouttmp{};

			//
			// This is the 4 channel representation of the output signal
			//
			Asample sigoutbuf[4];

			//
			// ofiltA/ofiltB do some removal of any resulting out of band stuff.
			// oddly -- the filters are actually much more effective near the
			// highest audio frequencies at the lower sample rates.  However,
			// I suspect that the decoder otherwise generally works best at 96k.
			// 192k requires twice the CPU, and I doubt that there is not
			// much benefit.
			// (at 44.1k samples/sec, the filter falls to -60dB by 21000,
			// while it sits fat and happy at about -35dB at 30kHz with
			// a 96k sample rate -- odd, for otherwise the same filter
			// parameters, heh?  Thats the world of digital filters.)
			//
			// I could have done a really good brickwall at 20500Hz, and just
			// might do so in the future -- but it does require lots of taps,
			// and might slow down the program a bit when running at higher
			// sample rates.  As is, the two filters are only 4th order IIR,
			// elliptical with a very little bit of ripple.  Since there are
			// two filters, I guess that the -35dB at 30kHz becomes -70dB,
			// and the middling frequencies of 25kHz are pretty well suppressed.
			// That is why I did two filters instead of one.  And also, back
			// when I thought that we needed to do the multi-axis version of
			// the expander, I didn't want subsequent steps to be interfered
			// with by the earlier steps.
			//
			// Also, and most importantly -- earlier versions of the code used
			// more real feedback of the actual audio -- and could fall into
			// oscillation if I made a bug, or even if the math on a processor
			// or library might work a little differently.  So, with so very
			// much attenuation at frequencies higher than might be noticed, it
			// might save a tweeter or some other physical device if the program
			// goes nuts.  It is very, very stable now.  The only thing that
			// could POSSIBLY go unstable is the gain control, not the audio
			// itself.
			//

			//
			// Here, we calculate the raw unfiltered signal squared
			// this is mostly just being used for absolute value
			// but in the past, this section of code came from a very
			// nice compressor which used carefully designed RMS
			// measurements.  Most code long gone -- I just do it
			// this way now.
			//
			vtype4 LRLINrmsinsql(VFP4(1.0));
			vtype4 LRLINrmsinsqr(VFP4(1.0));
			vtype4 M90LINrmsinsql(VFP4(1.0));
			vtype4 M90LINrmsinsqr(VFP4(1.0));
			vtype4 MSLINrmsinsql(VFP4(1.0));
			vtype4 MSLINrmsinsqr(VFP4(1.0));

			for (int i = 0; i < 4; i++)
			{
				Asample& LRstmp = siginLR[i];

				float lsig = ((LRstmp.item0() * LRstmp.item0()));
				LRLINrmsinsql[i] = lsig;
				float rsig = ((LRstmp.item1() * LRstmp.item1()));
				LRLINrmsinsqr[i] = rsig;

				if (mode & 0x1000)
				{
					Asample& M90stmp = sigin90[i];

					lsig = ((M90stmp.item0() * M90stmp.item0()));
					M90LINrmsinsql[i] = lsig;
					rsig = ((M90stmp.item1() * M90stmp.item1()));
					M90LINrmsinsqr[i] = rsig;
				}

				Asample& MSstmp = siginMS[i];

				lsig = ((MSstmp.item0() * MSstmp.item0()));
				MSLINrmsinsql[i] = lsig;
				rsig = ((MSstmp.item1() * MSstmp.item1()));
				MSLINrmsinsqr[i] = rsig;
			}

			//
			// All filters are linear (NOT SQUARED -- unlike most of my tech, where I tend to work
			// in the squared and dB domain.)
			//

			//
			// get the linear values
			// & add in the threshold as determined above
			//
			vtype4 LRLINrmsinl = vsqrt(LRLINrmsinsql);
			vtype4 LRLINrmsinr = vsqrt(LRLINrmsinsqr);
			LRLINrmsinl += INPthreshold;
			LRLINrmsinr += INPthreshold;

			vtype4 M90LINrmsinl{};
			vtype4 M90LINrmsinr{};
			if (mode & 0x1000)
			{
				M90LINrmsinl = vsqrt(M90LINrmsinsql);
				M90LINrmsinr = vsqrt(M90LINrmsinsqr);
				M90LINrmsinl += XINPthreshold;
				M90LINrmsinr += XINPthreshold;
			}

			vtype4 MSLINrmsinl = vsqrt(MSLINrmsinsql);
			vtype4 MSLINrmsinr = vsqrt(MSLINrmsinsqr);
			MSLINrmsinl += INPthreshold;
			MSLINrmsinr += INPthreshold;

#if 0
			fprintf(stderr, "LRLINin: "); LRLINrmsinl.fprint(stderr); fprintf(stderr,"\n");
			fprintf(stderr, "M90LINin: "); M90LINrmsinl.fprint(stderr); fprintf(stderr,"\n");
			fprintf(stderr, "MSLINin: "); MSLINrmsinl.fprint(stderr); fprintf(stderr,"\n");
#endif

			//
			//
			// we can choose from an additive threshold or a hard
			// threshold
			//

			//
			// Do the attack/release -- the LG filter does a fast attack, and
			// millisecond decay. The LNB filter does the slow attack/slow decay
			// These filters are put together in a contorted fashion.   The reason why
			// these filters are put together this way is based upon the DolbyA behavior.
			// If you change the code between 'DOLBYA FILTER CODE BEGIN/END', the
			// characteristics will not likely match a real DolbyA unit.  I don't normally
			// write my attack/decay filters like this, so there is a real reason for this
			// structure.
			//

// APPROX DOLBYA FILTER TIMING CODE BEGIN
			vtype4 fatmp{};

			vtype4 LRfbtmpl = LRdfiltlLG(LRLINrmsinl);
			fatmp = Vselect(Vgt(LRLINrmsinl, LRfbtmpl), (LRLINrmsinl + LRfbtmpl) * VFP4(0.50), LRfbtmpl);
			vtype4 LINrmsrawtmpl = LRdfiltlLNB(fatmp);
			LINrmsrawtmpl = Vselect(Vgt(fatmp, LINrmsrawtmpl), (fatmp + LINrmsrawtmpl) * VFP4(0.50), LINrmsrawtmpl);

			vtype4 LRfbtmpr = LRdfiltrLG(LRLINrmsinr);
			fatmp = Vselect(Vgt(LRLINrmsinr, LRfbtmpr), (LRLINrmsinr + LRfbtmpr) * VFP4(0.50), LRfbtmpr);
			vtype4 LINrmsrawtmpr = LRdfiltrLNB(fatmp);
			LINrmsrawtmpr = Vselect(Vgt(fatmp, LINrmsrawtmpr), (fatmp + LINrmsrawtmpr) * VFP4(0.50), LINrmsrawtmpr);

			vtype4 M90LINrmsrawtmpl{};
			vtype4 M90LINrmsrawtmpr{};
			vtype4 M90fbtmpl{};
			vtype4 M90fbtmpr{};

			if (mode & 0x1000)
			{
				M90fbtmpl = M90dfiltlLG(M90LINrmsinl);
				fatmp = Vselect(Vgt(M90LINrmsinl, M90fbtmpl), (M90LINrmsinl + M90fbtmpl) * VFP4(0.50), M90fbtmpl);
				M90LINrmsrawtmpl = M90dfiltlLNB(fatmp);
				M90LINrmsrawtmpl = Vselect(Vgt(fatmp, M90LINrmsrawtmpl), (fatmp + M90LINrmsrawtmpl) * VFP4(0.50), M90LINrmsrawtmpl);
			

				M90fbtmpr = M90dfiltrLG(M90LINrmsinr);
				fatmp = Vselect(Vgt(M90LINrmsinr, M90fbtmpr), (M90LINrmsinr + M90fbtmpr) * VFP4(0.50), M90fbtmpr);
				M90LINrmsrawtmpr = M90dfiltrLNB(fatmp);
				M90LINrmsrawtmpr = Vselect(Vgt(fatmp, M90LINrmsrawtmpr), (fatmp + M90LINrmsrawtmpr) * VFP4(0.50), M90LINrmsrawtmpr);
			}

			vtype4 MSfbtmpl = MSdfiltlLG(MSLINrmsinl);
			fatmp = Vselect(Vgt(MSLINrmsinl, MSfbtmpl), (MSLINrmsinl + MSfbtmpl) * VFP4(0.50), MSfbtmpl);
			vtype4 MSLINrmsrawtmpl = MSdfiltlLNB(fatmp);
			MSLINrmsrawtmpl = Vselect(Vgt(fatmp, MSLINrmsrawtmpl), (fatmp + MSLINrmsrawtmpl) * VFP4(0.50), MSLINrmsrawtmpl);

			vtype4 MSfbtmpr = MSdfiltrLG(MSLINrmsinr);
			fatmp = Vselect(Vgt(MSLINrmsinr, MSfbtmpr), (MSLINrmsinr + MSfbtmpr) * VFP4(0.50), MSfbtmpr);
			vtype4 MSLINrmsrawtmpr = MSdfiltrLNB(fatmp);
			MSLINrmsrawtmpr = Vselect(Vgt(fatmp, MSLINrmsrawtmpr), (fatmp + MSLINrmsrawtmpr) * VFP4(0.50), MSLINrmsrawtmpr);

// APPROX DOLBYA FILTER TIMING CODE END

			static constexpr double ssoffset = 0.0;
			LINrmsrawtmpl = Vmax(rfpl.next(LRfbtmpl), LINrmsrawtmpl * vtype4(ssoffset));
			LINrmsrawtmpr = Vmax(rfpr.next(LRfbtmpr), LINrmsrawtmpr * vtype4(ssoffset));
			if (mode & 0x1000)
			{
				M90LINrmsrawtmpl = Vmax(M90rfpl.next(M90fbtmpl), M90LINrmsrawtmpl * vtype4(ssoffset));
				M90LINrmsrawtmpr = Vmax(M90rfpr.next(M90fbtmpr), M90LINrmsrawtmpr * vtype4(ssoffset));
			}

			MSLINrmsrawtmpl = Vmax(MSrfpl.next(MSfbtmpl), MSLINrmsrawtmpl * vtype4(ssoffset));
			MSLINrmsrawtmpr = Vmax(MSrfpr.next(MSfbtmpr), MSLINrmsrawtmpr * vtype4(ssoffset));

			//
			// get the dB values for the dB calculations -- I derived my expansion curves
			// in the dB domain and haven't done the conversion to linear yet.
			// This is a technically unnecessary waste, but not too bad on recent machines.
			//
			// I have decided to remove the special threshold code (and support for the
			// differing LF/HF compression/expansion curve versions)
			//
			// All we need to do now is to convert the dB/dB gain curve to the
			// linear world.   Curve-fitting time again!!!  I dont' think that it is
			// possible to do the change algebraically (I already tried doing it myself
			// and using Maxima.)  I want to be fairly faithful to the curve -- it took
			// significant amount of time to tweak the curve until it sounded right.  If
			// I let myself, I could probably tweak it until the end of time.
			//
			// removing the lin-to-dB, applying the curve, then dB-to-lin, and just
			// doing a curve should greatly simplify the code and really speed it up.
			//
			vtype4 dBlvll = lin2dB(LINrmsrawtmpl);
			vtype4 dBlvlr = lin2dB(LINrmsrawtmpr);

			vtype4 M90dBlvll{};
			vtype4 M90dBlvlr{};
			if (mode & 0x1000)
			{
				M90dBlvll = lin2dB(M90LINrmsrawtmpl);
				M90dBlvlr = lin2dB(M90LINrmsrawtmpr);
			}

			vtype4 MSdBlvll = lin2dB(MSLINrmsrawtmpl);
			vtype4 MSdBlvlr = lin2dB(MSLINrmsrawtmpr);

#if 0
			fprintf(stderr, "LRdBlvl: "); dBlvll.fprint(stderr); fprintf(stderr,"\n");
			fprintf(stderr, "M90dBlvl: "); M90dBlvll.fprint(stderr); fprintf(stderr,"\n");
			fprintf(stderr, "MSdBlvl: "); MSdBlvll.fprint(stderr); fprintf(stderr,"\n");
#endif

			//
			// set the level to pivot against
			//
			dBlvll -= dBEXpvtI;
			dBlvlr -= dBEXpvtI;
			if (mode & 0x1000)
			{
				M90dBlvll -= dBEXpvtI;
				M90dBlvlr -= dBEXpvtI;
			}
			MSdBlvll -= dBEXpvtI;
			MSdBlvlr -= dBEXpvtI;

			//
			// Here, we apply the compression ratio (currently in dB terms)
			//
			vtype4 dBnll = fgfe(dBlvll);
			vtype4 dBnlr = fgfe(dBlvlr);

			vtype4 M90dBnll{};
			vtype4 M90dBnlr{};
			if (mode & 0x1000)
			{
				M90dBnll = fgfe(M90dBlvll);
				M90dBnlr = fgfe(M90dBlvlr);
			}
			vtype4 MSdBnll = fgfe(MSdBlvll);
			vtype4 MSdBnlr = fgfe(MSdBlvlr);

			
			dBnll = Vmin(dBnll, vtype4(-0.50));
			dBnlr = Vmin(dBnlr, vtype4(-0.50));
			if (mode & 0x1000)
			{
				M90dBnll = Vmin(M90dBnll, vtype4(-0.50));
				M90dBnlr = Vmin(M90dBnlr, vtype4(-0.50));
			}
			MSdBnll = Vmin(MSdBnll, vtype4(-0.50));
			MSdBnlr = Vmin(MSdBnlr, vtype4(-0.50));

			//
			// Convert back to linear
			//
			pgainl = dB2lin(dBnll);
			pgainr = dB2lin(dBnlr);
			if (mode & 0x1000)
			{
				M90pgainl = dB2lin(M90dBnll);
				M90pgainr = dB2lin(M90dBnlr);
			}

			MSpgainl = dB2lin(MSdBnll);
			MSpgainr = dB2lin(MSdBnlr);

			//
			// Here we move the level values around for various legacy reasons.  If you are trying to figure out
			// WHY I am moving the values around as much as I am -- don't bother.  I am just doing it because the
			// old code had reasons to do so.  Now, there is a serious need for a code cleanup!!!
			//
			vtype4 dgainl{}, MSdgainl{}, M90dgainl{};
			vtype4 dgainr{}, MSdgainr{}, M90dgainr{};

			dgainl = pgainl;
			dgainr = pgainr;
			if (mode & 0x1000)
			{
				M90dgainl = M90pgainl;
				M90dgainr = M90pgainr;
			}
			MSdgainl = MSpgainl;
			MSdgainr = MSpgainr;

			//
			// Slow down the gain changes a little (help to avoid aliasing) -- also avoid any errors
			// in secret sauce.
			//
			pgainl = dfiltlA(pgainl);
			pgainr = dfiltrA(pgainr);
			if (mode & 0x1000)
			{
				M90pgainl = M90dfiltlA(M90pgainl);
				M90pgainr = M90dfiltrA(M90pgainr);
			}
			MSpgainl = MSdfiltlA(MSpgainl);
			MSpgainr = MSdfiltrA(MSpgainr);

			//
			// Here, we are calculating the gain for 9k-20k
			//
			VFP4 pgainl3{}, M90pgainl3{}, MSpgainl3{};
			VFP4 pgainr3{}, M90pgainr3{}, MSpgainr3{};
			pgainl3 = (pgainl[3] + pgainl[2]) * VFP4(0.562);
			pgainr3 = (pgainr[3] + pgainr[2]) * VFP4(0.562);
			if (mode & 0x1000)
			{
				M90pgainl3 = (M90pgainl[3] + M90pgainl[2]) * VFP4(0.562);
				M90pgainr3 = (M90pgainr[3] + M90pgainr[2]) * VFP4(0.562);
			}
			MSpgainl3 = (MSpgainl[3] + MSpgainl[2]) * VFP4(0.562);
			MSpgainr3 = (MSpgainr[3] + MSpgainr[2]) * VFP4(0.562);

			//
			// Set the 10dB loss limit for the lower three bands
			//
			for (int i = 0; i < 3; i++)
			{
				pgainl[i] = Vmax(pgainl[i], VFP4(0.316));
				pgainr[i] = Vmax(pgainr[i], VFP4(0.316));
				if (mode & 0x1000)
				{
					M90pgainl[i] = Vmax(M90pgainl[i], VFP4(0.316));
					M90pgainr[i] = Vmax(M90pgainr[i], VFP4(0.316));
				}
				MSpgainl[i] = Vmax(MSpgainl[i], VFP4(0.316));
				MSpgainr[i] = Vmax(MSpgainr[i], VFP4(0.316));
			}

			//
			// set the 15dB loss limit for the highest band
			//
			pgainl[3] = Vmax(pgainl3, VFP4(0.316 * 0.562));
			pgainr[3] = Vmax(pgainr3, VFP4(0.316 * 0.562));
			if (mode & 0x1000)
			{
				M90pgainl[3] = Vmax(M90pgainl3, VFP4(0.316 * 0.562));
				M90pgainr[3] = Vmax(M90pgainr3, VFP4(0.316 * 0.562));
			}
			MSpgainl[3] = Vmax(MSpgainl3, VFP4(0.316 * 0.562));
			MSpgainr[3] = Vmax(MSpgainr3, VFP4(0.316 * 0.562));


			//
			// gain never higher than 0dB
			//
			pgainl = Vmin(pgainl, vtype4(VFP4(1.0)));
			pgainr = Vmin(pgainr, vtype4(VFP4(1.0)));
			if (mode & 0x1000)
			{
				M90pgainl = Vmin(M90pgainl, vtype4(VFP4(1.0)));
				M90pgainr = Vmin(M90pgainr, vtype4(VFP4(1.0)));
			}
			MSpgainl = Vmin(MSpgainl, vtype4(VFP4(1.0)));
			MSpgainr = Vmin(MSpgainr, vtype4(VFP4(1.0)));
					

			//
			// seperate 3-9k and 9-20k
			//
			Asample sigtmp2 = siginLR[2] - siginLR[3];

			//
			// calculate new signal levels for each of the 4 bands
			// 0-80, 80-3k, 3k-9k, 9k-20k
			//

			if (mode & 0x1)
			{
				sigin[0] = siginLR[0];
				sigin[1] = siginLR[1];
				pgainl[0] = 1.0f;
				pgainl[1] = 1.0f;
				pgainr[0] = 1.0f;
				pgainr[1] = 1.0f;
			}
			else
			{
				sigin[0] = Asample{ siginLR[0].item0() * float(pgainl[0]), siginLR[0].item1() * float(pgainr[0]) };
				sigin[1] = Asample{ siginLR[1].item0() * float(pgainl[1]), siginLR[1].item1() * float(pgainr[1]) };
			}
			sigin[2] = Asample{ sigtmp2.item0() * float(pgainl[2]), sigtmp2.item1() * float(pgainr[2]) };
			sigin[3] = Asample{ siginLR[3].item0() * float(pgainl[3]), siginLR[3].item1() * float(pgainr[3])};

			if (mode & 0x1000)
			{
				for (int i = 0; i < 4; i++)
				{
					sigin90[i] = ninetymatrix(sigin[i], false);
				}
				sigtmp2 = sigin90[2] - sigin90[3];

				if (mode & 0x1)
				{
					M90pgainl[0] = 1.0f;
					M90pgainl[1] = 1.0f;
					M90pgainr[0] = 1.0f;
					M90pgainr[1] = 1.0f;
				}
				else
				{
					sigin90[0] = Asample{ sigin90[0].item0() * float(M90pgainl[0]), sigin90[0].item1() * float(M90pgainr[0]) };
					sigin90[1] = Asample{ sigin90[1].item0() * float(M90pgainl[1]), sigin90[1].item1() * float(M90pgainr[1]) };
				}
				sigin90[2] = Asample{ sigtmp2.item0() * float(M90pgainl[2]), sigtmp2.item1() * float(M90pgainr[2]) };
				sigin90[3] = Asample{ sigin90[3].item0() * float(M90pgainl[3]), sigin90[3].item1() * float(M90pgainr[3])};

				for (int i = 0; i < 4; i++)
				{
					sigin[i] = ninetymatrix(sigin90[i], true);
				}
			}

			for (int i = 0; i < 4; i++)
			{
				siginMS[i] = msmatrix(sigin[i], false);
			}
			sigtmp2 = siginMS[2] - siginMS[3];

			if (mode & 0x1)
			{
				MSpgainl[0] = 1.0f;
				MSpgainl[1] = 1.0f;
				MSpgainr[0] = 1.0f;
				MSpgainr[1] = 1.0f;
			}
			else
			{
				siginMS[0] = Asample{ siginMS[0].item0() * float(MSpgainl[0]), siginMS[0].item1() * float(MSpgainr[0]) };
				siginMS[1] = Asample{ siginMS[1].item0() * float(MSpgainl[1]), siginMS[1].item1() * float(MSpgainr[1]) };
			}
			siginMS[2] = Asample{ sigtmp2.item0() * float(MSpgainl[2]), sigtmp2.item1() * float(MSpgainr[2]) };
			siginMS[3] = Asample{ siginMS[3].item0() * float(MSpgainl[3]), siginMS[3].item1() * float(MSpgainr[3])};

			for (int i = 0; i < 4; i++)
			{
				sigoutbuf[i] = msmatrix(siginMS[i], true);
			}

			Asample ob[4];

			if (( mode & 0x2) == 0)
			{
				outputfilt0(sigoutbuf, ob);
				outputfilt1(ob, ob);
			}
			else
			{
				ob[0] = sigoutbuf[0];
				ob[1] = sigoutbuf[1];
				ob[2] = sigoutbuf[2];
				ob[3] = sigoutbuf[3];
			}


			//
			// build the output -- summing each of the 4 bands
			//
			if (mode & 0x100)
			{
				for (int i = 1; i < 4; i++)
				{
					sigouttmp += ob[i];
				}
			}
			else if (mode & 0x200)
			{
				for (int i = 0; i < 1; i++)
				{
					sigouttmp += ob[i];
				}
			}
			else
			{
				for (int i = 0; i < 4; i++)
				{
					sigouttmp += ob[i];
				}
			}

			sigout = ofiltA(ofiltB(sigouttmp));

			prevsample = sigouttmp;

			return Xnextlevelrtval{
				(pgainl + pgainr) * VFP4(0.50),
				(M90pgainl + M90pgainr) * VFP4(0.50),
				(MSpgainl + MSpgainr) * VFP4(0.50)};

//* M90pgainl * MSpgainl + pgainr * M90pgainr * MSpgainr) * vtype4(VFP4(0.50));

			//return avggain;
	}

	void
	next(const Asample& sigin, Asample& sigout)
	{
			vtype4 gain = nextlevel(sigin, sigout);

			//
			// looking at the schematic -- the fet has a big resistor in the gate -- enough
			// that it should be accounted for by a little bit of a TC -- I used 1msec.
			//

			DLBYAst[myindex].next(gain);

			scnt++;
			return;
	}

	void
	Xnext(const Asample& sigin, Asample& sigout)
	{
			Xnextlevelrtval gain = Xnextlevel(sigin, sigout);

			//
			// looking at the schematic -- the fet has a big resistor in the gate -- enough
			// that it should be accounted for by a little bit of a TC -- I used 1msec.
			//

			//gain.v0.fprint(stderr);
			DLBYAst[myindex].next(gain.v0);
			DLBYAst[myindex+1].next(gain.v1);
			DLBYAst[myindex+2].next(gain.v2);

			scnt++;
			return;
	}

};
};


//
// This owns and runs the expansion passes -- traversing the control table --
// interpreting it and invoking the passes.
//
class expA
{
	static constexpr int NPASSES = MAXEXP;
	static constexpr int NFRQCHAN = 4;
	static constexpr float gaincomp = 1.000f;
	static constexpr float hfcomp = 0.000f;

	da::dbexpand dbx[NPASSES];
	//
	// This filter is meant to remove above-audio frequencies.  We can encounter >20kHz
	// because this program works up to 192ksamples/sec.
	//
	lpf20500 ifilt;

	Asample
	expstep(int index, const Asample& insamplearg)
	{
		Asample insample(insamplearg);

		insample *= expcfg[index].LINgainbefore;

		Asample outsample{};
		if (expcfg[index].dmode == DMODE_X)
			dbx[index].Xnext(insample, outsample);
		else
			dbx[index].next(insample, outsample);

		outsample *= gaincomp;
		return outsample;
	}

public:
	expA() : dbx(), ifilt()
	{
			//
			// set the indices for each of the expansion passes
			//
			for (int i = 0; i < NPASSES; i++)
			{
				dbx[i].setindex(i);
				dbx[i].settindex(i);
			}
	}

	//
	// this routine does all of the passes for each input sample
	//
	int // NOINLINE
	next(int count, const Asample *sigin, Asample *sigout)
	{
		for (int i = 0; i < count; i++)
		{
			Asample insample = ifilt(sigin[i]);
			Asample insampletmp{};
			int j = 0;
			for (j = 0; j < numexp; j++)
			{
				Asample originsample = insample;
				float mid, side;
				expcfgT& cfg = expcfg[j];
				float gbefore = cfg.LINgainbefore;
				switch (cfg.dmode)
				{

		case DMODE_SKIP:
					break;
		case DMODE_LR:
					insample = expstep(j, insample);
					break;

		case DMODE_LRGAINONLY:
					insample *= gbefore;
					DLBYAst[j].next(gbefore);
					break;

		case DMODE_LRWONLY:
					insample.setitem1(insample.item1() * gbefore);
					DLBYAst[j].next(gbefore);
					break;
					
		case DMODE_X:
					insample = expstep(j, insample);
					j += 2;
					break;

		case DMODE_MSGAINONLY:
		case DMODE_MSWONLY:
		case DMODE_MS:
					mid = insample.item0() + insample.item1();
					mid *= 0.707f;
					side = insample.item0() - insample.item1();
					side *= 0.707f;
					insampletmp = Asample{mid, side};

					if (cfg.dmode == DMODE_MS)
					{
						insampletmp = expstep(j, insampletmp);
					}
					else if (cfg.dmode == DMODE_MSWONLY)
					{
						insampletmp.setitem1(insampletmp.item1() * gbefore);
						DLBYAst[j].next(gbefore);
					}
					else
					{
						insampletmp *= gbefore;
						DLBYAst[j].next(gbefore);
					}

					mid = insampletmp.item0() + insampletmp.item1();
					mid *= 0.707f;
					side = insampletmp.item0() - insampletmp.item1();
					side *= 0.707f;

					//insample = (originsample * 0.50f + Asample{mid, side} * 0.50f);
					insample = Asample{mid, side};
					break;
		case DMODE_90GAINONLY:
		case DMODE_90WONLY:
		case DMODE_90:
					insampletmp = ninetymatrix(insample, false);
					if (cfg.dmode == DMODE_90)
					{
						insampletmp = expstep(j, insampletmp);
					}
					else if (cfg.dmode == DMODE_90WONLY)
					{
						insampletmp.setitem1(insampletmp.item1() * gbefore);
						DLBYAst[j].next(gbefore);
					}
					else
					{
						insampletmp *= gbefore;
						DLBYAst[j].next(gbefore);
					}
					//insample = (originsample * 0.50f + ninetymatrix(insampletmp, true) * 0.50f);
					insample = ninetymatrix(insampletmp, true);
					break;
				}
			}
			sigout[i] = insample;
		}

		return count;
	}

};

//
// This routine is extreme hackery -- I folded the infrastructure for the audio processor into one thread -- and cobbled one of the
// threads into this monstrosity (use of static variables/etc.)
//
// Should change this into a class -- but no time right now.
//
void
procmusic(paudiochunk pa)
{
		//
		// This 'static' stuff is cr*p, but it works for now.
		//
		static int totalsamples70;
		static int limit99, limit70, limit50, limit25;
		static bool isinit = false;

		static expA* exp = nullptr;

		if (!isinit)
		{
			exp = new expA;
			isinit = true;
		}

		audiopayload* ap = &pa->payload();
		pAsample sp = &ap->samples[0];
		int totalsamples = ap->samplecount;

		for (int i = 0; i < totalsamples; i++)
		{
			const Asample& as = ap->samples[i];
			float sumsq = as.msumsq();
			INst.next( sumsq);
		}

		exp->next(totalsamples, sp, sp);

		for (int i = 0; i < totalsamples; i++)
		{
			const Asample& as = ap->samples[i];
			float sumsq = as.msumsq();
			OUTst.next( sumsq);
			accuminfo::accumulatelevel(as, totalsamples70, limit99, limit70, limit50, limit25);
		}
}

//
// vi: set ts=3:
//

