//
// Recusive expansion control signal correction class
//

#if !defined(__RECURSEXP_H__)
#define __RECURSEXP_H__

#define AI __attribute__((always_inline))

namespace recursexp_h
{
	static constexpr int MAXLOOKUP = 64;
	//
	// fast lookup table for inverting integers
	//
#if defined(EIGHTWIDE)
	static constexpr double flookup[MAXLOOKUP] = {
#else
	static constexpr float flookup[MAXLOOKUP] = {
#endif
		1.0, -0.50, 1.0/3.0, -1.0/4.0, 1.0/5.0,
		-1.0/6.0, 1.0/7.0, -1.0/8.0, 1.0/9.0, -1.0/10.0,
		1.0/11.0, -1.0/12.0, 1.0/13.0, -1.0/14.0, 1.0/15.0,
		-1.0/16.0, 1.0/17.0, -1.0/18.0, 1.0/19.0, -1.0/20.0,
		1.0/21.0, -1.0/22.0, 1.0/23.0, -1.0/24.0, 1.0/25.0,
		-1.0/26.0, 1.0/27.0, -1.0/28.0, 1.0/29.0, -1.0/30.0,
		1.0/31.0, -1.0/32.0,
		1.0/33.0, -1.0/34.0, 1.0/35.0, -1.0/36.0,
		1.0/37.0, -1.0/39.0, 1.0/40.0, -1.0/41.0, 
		1.0/42.0, -1.0/43.0, 1.0/44.0, -1.0/45.0, 
		1.0/46.0, -1.0/47.0, 1.0/48.0, -1.0/49.0, 
		1.0/50.0, -1.0/51.0, 1.0/52.0, -1.0/53.0, 
		1.0/54.0, -1.0/55.0, 1.0/56.0, -1.0/57.0, 
		1.0/58,0, -1.0/59.0, 1.0/60.0, -1.0/61.0, 
		1.0/62,0, -1.0/63.0
	};


#if defined(EIGHTWIDE)
static constexpr v4df lpf0atk = (v4df) {satk, satk, fatk, fatk};
static constexpr v4df lpf0dcy = (v4df) {sdcy, sdcy, fdcy, fdcy};
#else
static constexpr v4sf lpf0atk = (v4sf) {satk, satk, fatk, fatk};
static constexpr v4sf lpf0dcy = (v4sf) {sdcy, sdcy, fdcy, fdcy};
#endif



	//
	// simple high pass filter class
	//
	template <typename VDEF>
	class LLhpf
	{
		using VT = varrayT<VDEF>;
		using IVT = IvarrayT<VDEF>;
      using FT = typename VDEF::FT;
      int nFT = VDEF::nFT;
	
      VT coeff;
      VT zval;

   public:
      LLhpf(void) AI : coeff(), zval(FT(0.00)) {}

      LLhpf(const LLhpf<VDEF>& init) AI : coeff(init.coeff), zval(init.zval) {}

      LLhpf(VT c, VT z = VT{FT(0.01)}) AI : coeff(c), zval(z) {}
	
      LLhpf(FT c, FT z = FT(0.01)) AI : LLhpf(VT(c), VT(z)) {}

      LLhpf<VDEF>&
      operator=(LLhpf<VDEF> src)
      {
         coeff = src.coeff;
         zval = src.zval;
         return *this;
      }

      VT AI
      next(const VT& in)
      {
#if 0
         VT out = in - zval;

         zval = zval + out * coeff;

         return out;
#else
			return in;
#endif
      }

      VT AI
      operator()(VT in)
      {
         return next(in);
   	}

   	void AI
   	setcoeff(VT newcoeff)
   	{
   		coeff = newcoeff;
   	}

   	void AI
   	setcoeff(FT newcoeff)
     	{
   		coeff = VT(newcoeff);
   	}
   };

   template <typename VDEF, int LEVEL = 0>
   class LLex0
   {
   	using VT = varrayT<VDEF>;
   	using IVT = IvarrayT<VDEF>;
   	using FT = typename VDEF::FT;
   	int nFT = VDEF::nFT;
   	using LPFA = Vtavfilt<VDEF>;

   	LPFA lpfa;

   public:

   	LLex0(void) AI :
			lpfa(sratetmod(VT(lpf0dcy)), (LEVEL == 0) ? VT(1.0f) : sratetmod(VT(lpf0atk)), VT(0.01f))
   	{}

   	VT AI
   	next(const VT& in)
   	{
			return lpfa(in);
   	}

		void
		setcoeffs(VT atka, VT dcya)
		{
			lpfa.setfilt(dcya, atka);
		}

		void
		setcoeffatk(VT atka)
		{
			lpfa.setavfilt(atka);
		}

		void
		setcoeffdcy(VT dcya)
		{
			lpfa.setfilt(dcya);
		}

   	VT AI
   	operator()(const VT& in)
   	{
   		return next(in);
   	}

   };

	template <typename VDEF>
	class recurseexpcorBT
	{
   	using VT = varrayT<VDEF>;
   	using IVT = IvarrayT<VDEF>;
   	using FT = typename VDEF::FT;
   	int nFT = VDEF::nFT;

	public:
		recurseexpcorBT(void) { }

		virtual ~recurseexpcorBT(void) {}

		virtual VT next(const VT& in) = 0;

		virtual VT nextALL(const VT& in) = 0;

		virtual VT operator() (const VT& in) = 0;

	};
		

   template <typename VDEF, typename FILT>
   class recurseexpcorTNIL : public recurseexpcorBT<VDEF>
   {
   	using VT = varrayT<VDEF>;
public:
   	recurseexpcorTNIL(void) { }

   	VT AI
   	next(const VT& in)
   	{
			return in;
   	}

   	VT AI
   	nextALL(const VT& in)
   	{
   		return next(in);
   	}

   	VT AI
		operator() (const VT& in)
   	{
   		return next(in);
   	}

   };


   template <typename VDEF, typename FILT, typename LOWEREXP, int MAXWIDTH, int MAXFULLWIDTH, int RLEVEL = 0>
   class recurseexpcorT : public recurseexpcorBT<VDEF>
   {
   	using VT = varrayT<VDEF>;
   	using IVT = IvarrayT<VDEF>;
   	using FT = typename VDEF::FT;
   	int nFT = VDEF::nFT;

   	LOWEREXP lower[MAXWIDTH];
   	FILT filt[MAXWIDTH];

public:
		recurseexpcorT(void) AI : lower(), filt() 
		{
			filt[0].setcoeffatk(VT(FT(1.0)));
		}

   	VT AI
   	nextALL(const VT& in)
   	{

			VT fval = in;
			VT fvalA = in;
			static constexpr int iincr = 1;
			int i = 0;
			VT addval(FT(0.0));

			int lindex = 0;
			VT addinc{};

			for ( ; i < MAXFULLWIDTH; i += iincr)
			{
				
				VT lowerseries = lower[lindex].nextALL(fval);
				VT lowerfilt = filt[lindex](lowerseries);
				addinc = lowerfilt * VT(flookup[i]);
				addval += addinc;

				fval *= fvalA;

				lindex += 1;
			}

   		for (; i < MAXWIDTH; i += iincr)
   		{
				VT lowerseries = fval;
				VT lowerfilt = filt[lindex](lowerseries);
				addinc = lowerfilt * VT(flookup[i]);
				addval += addinc;

				fval *= fvalA;

				lindex += 1;
			}

   		VT rtval = (addval - addinc * VT(FT(0.50)));
			rtval = Vselect(Vgt(in, VT(FT(1.0))), in, rtval);
   		return rtval;
   	}

   	VT NOINLINE
   	next(const VT& in)
   	{
   		return nextALL(in);
   	}

   	VT AI
		operator() (const VT& in)
   	{
   		return next(in);
   	}

   };
   using rfiltVDEF = VDEF4;

   static constexpr int MAXWIDTHL1V0 = 11;
   static constexpr int MAXWIDTHL1V0F = 7;
   static constexpr int MAXWIDTHL0V0 = 31;
   static constexpr int MAXWIDTHL0V0F = 13;
   using V0recursefiltNIL = recurseexpcorTNIL<rfiltVDEF, LLex0<rfiltVDEF,2> >;
   using V0recursefiltl1 = recurseexpcorT<rfiltVDEF, LLex0<rfiltVDEF, 1>, V0recursefiltNIL, MAXWIDTHL1V0, MAXWIDTHL1V0F, 1>;
   using V0recursefiltl0 = recurseexpcorT<rfiltVDEF, LLex0<rfiltVDEF, 0>, V0recursefiltl1, MAXWIDTHL0V0, MAXWIDTHL0V0F, 0>;

   static constexpr int MAXWIDTHL2V1 = 5;
   static constexpr int MAXWIDTHL2V1F = 3;
   static constexpr int MAXWIDTHL1V1 = 9;
   static constexpr int MAXWIDTHL1V1F = 7;
   static constexpr int MAXWIDTHL0V1 = 35;
   static constexpr int MAXWIDTHL0V1F = 11;
   using V1recursefiltNIL = recurseexpcorTNIL<rfiltVDEF, LLex0<rfiltVDEF,3> >;
   using V1recursefiltl2 = recurseexpcorT<rfiltVDEF, LLex0<rfiltVDEF, 2>, V1recursefiltNIL, MAXWIDTHL2V1, MAXWIDTHL2V1F, 2>;
   using V1recursefiltl1 = recurseexpcorT<rfiltVDEF, LLex0<rfiltVDEF, 1>, V1recursefiltl2, MAXWIDTHL1V1, MAXWIDTHL1V1F, 1>;
   using V1recursefiltl0 = recurseexpcorT<rfiltVDEF, LLex0<rfiltVDEF, 0>, V1recursefiltl1, MAXWIDTHL0V1, MAXWIDTHL0V1F, 0>;

};

//using V0recursefilt = recursexp_h::V0recursefiltl0;
using V0recursefilt = recursexp_h::V1recursefiltl0;
using recursefilt = recursexp_h::recurseexpcorBT<recursexp_h::rfiltVDEF>;


#endif

//
// vi: set ts=3:
//
