#include <stdio.h>
#include <string.h>

#if !defined(offsetof)
#define offsetof __builtin_offsetof
#endif

static const int wav0_nchannels = 2;
static const int wav0_bits_per_sample = 16;

struct __attribute__((packed)) tag
{
	char cdata[4];

	tag(void) : cdata() {}

	tag(const char *tn) : cdata()
	{
		strncpy(cdata, tn, 4);
	}
};

struct tagsize
{
	tag name;
	int len;

	tagsize() : name(), len() {}
};

struct rwav0 {
	tag	riffsn;
	int	rlength;
	tag	wavsn;
};


struct __attribute__((packed)) wavh {

	tag	riffsn;
	int	rlength;
	tag	wavsn;
	tag	fmtsn;
	int	flength;
	short	aformat;
	short	nchannels;
	int	srate;
	int	byrate;
	short	balign;
	short	bps;
	short	extra;
	tag	factsn;
	int	factlen;
	int	factdata;
	tag	datasn;
	int	datalen;

	wavh(int wavtype = 1, int asrate = baseSRATE) :
		riffsn("RIFF"), rlength(0), wavsn("WAVE"), fmtsn("fmt "),
		flength(18), aformat(1), nchannels(wav0_nchannels), srate(asrate),
		byrate(wav0_nchannels*asrate*wav0_bits_per_sample/8),
		balign(wav0_nchannels*wav0_bits_per_sample/8),
		bps(wav0_bits_per_sample), extra(0),
		factsn("fact"), factlen(4), factdata(0), datasn("data"), datalen(0)
	{
		//
		// Default above is signed integer 16 bit, type 3 is fp 32bit
		if (wavtype == 3)
		{
			aformat = 3;
			bps = 32;
			byrate = nchannels * srate * (bps / 8);
			balign = nchannels * (bps / 8);
		}
	}

// header type: fmt , size: 18 found
// header type: fact, size: 4 found
// Wave file data found at 58, size: 153904344, fmt at: 20, wav header:
// riffsn: "RIFF", rlength: 153904394, wavsn: "WAVE", fmtsn: "RIFF", flength: 153904394
//   audioformat: 3, nchannels: 2, srate: 96000, byrate: 768000, balign: 8
//   bps: 32, datasn: "    ", datalen: 1996488706

	void
	setfsize(int length)
	{
		int dlength;
		dlength = length - sizeof(wavh);
		datalen = dlength;

		rlength = length - 8;
	}


	void
	snprint(char *buf, int bufsize)
	{
		snprintf(buf, bufsize,
			"riffsn: \"%4.4s\", rlength: %d, wavsn: \"%4.4s\", fmtsn: \"%4.4s\", flength: %d\n"
			"  audioformat: %d, nchannels: %d, srate: %d, byrate: %d, balign: %d\n"
			"  bps: %d, extra: %d, datasn: \"%4.4s\", datalen: %d\n",
			riffsn.cdata, rlength, wavsn.cdata, fmtsn.cdata, flength,
			aformat, nchannels, srate, byrate,
			balign, bps, extra, datasn.cdata, datalen);
	}
};



//
// Find the audio data in a wav file.  Return the basic wav header
// in the structure pointed to by the 2nd arg.  Return the FP
// state pointing to the audio data, and also the location of the
// audio data is returned.  If an error, then return -1ll.
//
static int __attribute__((unused))
finddata(FILE *fp, char *rthdr, int hdrmaxsize, int *fmtptr, int *datasize = nullptr)
{
	char tbuf[16384];
	char rawbuf[16384];
	int rawbptr = 0;
//
// rewind the file (Ignored for pipes)
//
	fseek(fp, 0L, SEEK_SET);

//
// read the file header
//
	int nbr = fread(tbuf, 1, sizeof (rwav0), fp);

	if (nbr < ((int) sizeof (rwav0)))
	{
		fprintf(stderr, "finddata: file too small for a wav file\n");
		return -1L;
	}

	memcpy(&rawbuf[rawbptr], tbuf, nbr);
	rawbptr += nbr;

	//
	// First part of wav header
	//

	if (!memcmp(&rawbuf[offsetof(wavh, riffsn)], "OggS", 4))
	{
		fprintf(stderr, "wav file type is OggS -- error\n");
		return -2;
	}

	if (memcmp(&rawbuf[offsetof(wavh, riffsn)], "RIFF", 4))
	{
		fprintf(stderr, "wav file type isn't riff -- error\n");
		return -1;
	}

	if (memcmp(&rawbuf[offsetof(wavh, wavsn)], "WAVE", 4))
	{
		fprintf(stderr, "RIFF file type is not \"WAVE\" -- error\n");
		return -1;
	}

	tagsize ts;
	while (true)
	{
		if (( nbr = fread(&ts, 1, sizeof ts, fp)) != sizeof ts)
		{
			fprintf(stderr, "Cannot read data tag in wav file\n");
			return -1;
		}

		memcpy(&rawbuf[rawbptr], &ts,  nbr);
		rawbptr += nbr;

		// fprintf(stderr,"header type: %4.4s, size: %d found\n", ts.name.cdata, ts.len);

		if (!memcmp(ts.name.cdata, "fmt ", 4))
		{
			*fmtptr = rawbptr;
		}
		else if (!memcmp(ts.name.cdata, "data", 4))
		{
			if (datasize)
				*datasize = ts.len;
			break;
		}

		//
		// Hack alert below
		//
		if (ts.len == 0 || ts.len > 1024*1024*1024)
		{
			//fprintf(stderr,"probable data found: \"%4.4s\"\n", ts.name.cdata);
			*datasize = -1;
			break;
		}


		int needtoread = ts.len;

		bool stopcopy = false;
		while (needtoread > 0)
		{
			int thisreadi = (needtoread > ((int)sizeof tbuf)) ? sizeof tbuf : needtoread;

			nbr = fread(tbuf, 1, thisreadi, fp);
			if (nbr != thisreadi)
			{
				fprintf(stderr, "Data read skipped off end of wav file -- error\n");
				return -1;
			}

			if (!stopcopy)
			{
				if ((rawbptr + thisreadi) > ((int) sizeof rawbuf))
				{
					fprintf(stderr, "Not able to save wave header -- too big\n");
					stopcopy = true;
				}
				else
				{
					memcpy(&rawbuf[rawbptr], tbuf, thisreadi);
				}
			}

			rawbptr += thisreadi;
			needtoread -= thisreadi;
		}
	}

	if (rthdr)
	{
		memcpy(rthdr, rawbuf, rawbptr);
		if (rawbptr > hdrmaxsize)
		{
			fprintf(stderr,"Warning -- header is truncated by hdrmaxsize(%d) too small, needs: %d\n",
				hdrmaxsize, rawbptr);
		}
	}
	return rawbptr;
}


